from .._content_type import ContentType
from ...delivery._data._data_provider import DataProviderLayer


class Definition(DataProviderLayer):
    """
    This class describe parameters to retrieve data for search custom instrument

    Parameters
    ----------
    access : str
        The search based on relationship to the custom instrument, for now only "owner" is supported. Can be omitted, default value is "owner"

    Examples
    --------
    >>> from refinitiv.data.content.custom_instruments import search
    >>> definition_search = search.Definition("VOD.L")
    >>> response = definition_search.get_data()
    """

    def __init__(self, access: str = "owner"):
        super().__init__(data_type=ContentType.CUSTOM_INSTRUMENTS_SEARCH, access=access)



from ._file_downloader import _FileDownloader
from ._unpacker import _Unpacker
from ... import _log as log
from ..._tools._common import cached_property


def logger():
    return log.root_logger().getChild("file-downloader")


class FileDownloader:
    def __init__(self, url, filename_ext) -> None:
        self._url = url
        self._filename_ext = filename_ext
        self._downloaded_filepath = None

    @cached_property
    def _downloader(self) -> _FileDownloader:
        return _FileDownloader(logger())

    @cached_property
    def _unpacker(self) -> _Unpacker:
        return _Unpacker(logger())

    def download(self, path="") -> "FileDownloader":
        self._downloaded_filepath = self._downloader.download(
            self._url, self._filename_ext, path
        )
        return self

    def extract(self, path="") -> str:
        filepath = self._downloaded_filepath or self._filename_ext
        extracted_filepath = self._unpacker.unpack(filepath, path)
        return extracted_filepath



__all__ = ("Definition", "events", "summaries", "search", "manage", "Intervals")
from ._definition import Definition
from . import events, summaries, search, manage
from .._intervals import Intervals



# coding: utf8
# contract_gen 2020-05-18 08:30:59.245816

__all__ = ["IndexResetType"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class IndexResetType(Enum):
    IN_ADVANCE = "InAdvance"
    IN_ARREARS = "InArrears"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(IndexResetType, _INDEX_RESET_TYPE_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_INDEX_RESET_TYPE_VALUES_IN_LOWER_BY_INDEX_RESET_TYPE, some)


_INDEX_RESET_TYPE_VALUES = tuple(t.value for t in IndexResetType)
_INDEX_RESET_TYPE_VALUES_IN_LOWER_BY_INDEX_RESET_TYPE = {
    name.lower(): item for name, item in IndexResetType.__members__.items()
}



# coding: utf-8

from pandas import DataFrame

from ....delivery._data._data_provider import (
    DataProvider,
    RequestFactory,
)

# ---------------------------------------------------------------------------
#   Content data
# ---------------------------------------------------------------------------

_response_universe_name = "universe"
_response_ric_name = "ric"
_response_display_name_name = "displayName"
_response_service_name_name = "serviceName"
_response_data_name = "data"
_response_constituents_name = "constituents"


def chains_build_df(raw, **kwargs):
    universe = raw[_response_universe_name]
    ric = universe[_response_ric_name]
    data = raw[_response_data_name]
    constituents = data[_response_constituents_name]
    _df = None
    if len(constituents):
        _df = DataFrame({ric: constituents})
    else:
        _df = DataFrame([], columns=[ric])
    if not _df.empty:
        _df = _df.convert_dtypes()
    return _df


# ---------------------------------------------------------------------------
#   Request factory
# ---------------------------------------------------------------------------


class ChainsRequestFactory(RequestFactory):
    def get_url(self, *args, **kwargs):
        url = args[1]
        url = url + "?universe={universe}"
        return url

    def get_path_parameters(self, *_, **kwargs):
        universe = kwargs.get("universe")
        if universe is None:
            return {}
        return {"universe": universe}


# ---------------------------------------------------------------------------
#   Data provider
# ---------------------------------------------------------------------------

chains_data_provider = DataProvider(
    request=ChainsRequestFactory(),
)



# coding: utf8
# contract_gen 2020-05-18 08:30:59.248815

__all__ = ["SwaptionSettlementType"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class SwaptionSettlementType(Enum):
    """
    The settlement type of the option if the option is exercised

    Possible values:
        - Cash
        - Physical
    """

    CASH = "Cash"
    PHYSICAL = "Physical"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(
            SwaptionSettlementType, _SWAPTION_SETTLEMENT_TYPE_VALUES, some
        )

    @staticmethod
    def normalize(some):
        return _normalize(
            _SWAPTION_SETTLEMENT_TYPE_VALUES_IN_LOWER_BY_SWAPTION_SETTLEMENT_TYPE, some
        )


_SWAPTION_SETTLEMENT_TYPE_VALUES = tuple(t.value for t in SwaptionSettlementType)
_SWAPTION_SETTLEMENT_TYPE_VALUES_IN_LOWER_BY_SWAPTION_SETTLEMENT_TYPE = {
    name.lower(): item for name, item in SwaptionSettlementType.__members__.items()
}



# coding: utf8

from typing import Optional

from .._base import UnderlyingDefinition


class FxUnderlyingDefinition(UnderlyingDefinition):
    """
    Parameters
    ----------
    fx_cross_code : str, optional
        The currency pair. Should contain the two currencies, ex EURUSD
    """

    def __init__(
        self,
        fx_cross_code: Optional[str] = None,
    ) -> None:
        super().__init__()
        self.fx_cross_code = fx_cross_code

    @property
    def fx_cross_code(self):
        """
        The currency pair. Should contain the two currencies, ex EURUSD
        :return: str
        """
        return self._get_parameter("fxCrossCode")

    @fx_cross_code.setter
    def fx_cross_code(self, value):
        self._set_parameter("fxCrossCode", value)



class RDError(Exception):
    """Base class for exceptions in this refinitiv-data module.

    Parameters
    ----------
    code: int
        error code for this exception
    message: str
        error description for this exception
    """

    def __init__(self, code: int, message: str):
        self.code = code
        self.message = message

    def __str__(self) -> str:
        return f"Error code {self.code} | {self.message}"


class SessionError(RDError):
    pass


class StreamingError(RDError):
    pass


class StreamConnectionError(RDError):
    pass


class PlatformSessionError(SessionError):
    pass


class DesktopSessionError(SessionError):
    pass


class ESGError(RDError):
    pass


class NewsHeadlinesError(RDError):
    pass


class EndpointError(RDError):
    pass


class StreamError(RDError):
    pass


class StreamingPricesError(RDError):
    pass


class ItemWasNotRequested(RDError):
    def __init__(self, item_type, not_requested, requested):
        item_type = f"{item_type}{'s' if len(not_requested) > 1 else ''}"
        super().__init__(
            -1, f"{item_type} {not_requested} was not requested : {requested}"
        )


class EnvError(RDError):
    pass


class BondPricingError(RDError):
    pass


class FinancialContractsError(RDError):
    pass


class RequiredError(RDError):
    pass


class ConfigurationError(RDError):
    pass



# coding: utf8


class CurvePoint:
    def __init__(self, x, y) -> None:
        self.x = x
        self.y = y

    def __str__(self) -> str:
        return f"x={self.x}, y={self.y}"

    def __repr__(self):
        return f"CurvePoint(x={self.x}, y={self.y})"



# coding: utf8

from typing import Optional

from ._eti_barrier_definition import EtiBarrierDefinition
from ..._instrument_definition import ObjectDefinition


class EtiDoubleBarriersDefinition(ObjectDefinition):
    """
    Parameters
    ----------
    barriers_definition : EtiBarrierDefinition, optional

    """

    def __init__(
        self,
        barriers_definition: Optional[EtiBarrierDefinition] = None,
    ) -> None:
        super().__init__()
        self.barriers_definition = barriers_definition

    @property
    def barriers_definition(self):
        """
        :return: list EtiBarrierDefinition
        """
        return self._get_list_parameter(EtiBarrierDefinition, "barriersDefinition")

    @barriers_definition.setter
    def barriers_definition(self, value):
        self._set_list_parameter(EtiBarrierDefinition, "barriersDefinition", value)



# coding: utf8
# contract_gen 2020-05-18 08:30:59.233816

__all__ = ["RepoCurveType"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class RepoCurveType(Enum):
    """
    Curve used to compute the repo rate.

    It can be computed using following methods:
        - DepositCurve : rate is computed by interpolating a deposit curve.
        - LiborFixing : rate is computed by interpolating libor rates.
        - RepoCurve : rate is computed by interpolating a repo curve.
    """

    DEPOSIT_CURVE = "DepositCurve"
    LIBOR_FIXING = "LiborFixing"
    REPO_CURVE = "RepoCurve"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(RepoCurveType, _REPO_CURVE_TYPE_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_REPO_CURVE_TYPE_VALUES_IN_LOWER_BY_REPO_CURVE_TYPE, some)


_REPO_CURVE_TYPE_VALUES = tuple(t.value for t in RepoCurveType)
_REPO_CURVE_TYPE_VALUES_IN_LOWER_BY_REPO_CURVE_TYPE = {
    name.lower(): item for name, item in RepoCurveType.__members__.items()
}



__all__ = (
    "full_measures",
    "full_scores",
    "basic_overview",
    "standard_measures",
    "standard_scores",
    "universe",
    "bulk",
)

from . import (
    full_measures,
    full_scores,
    basic_overview,
    standard_measures,
    standard_scores,
    universe,
    bulk,
)



# coding: utf-8
__all__ = (
    "Definition",
    "lookup",
    "metadata",
    "SearchViews",
)

from ._search_templates import templates as _templates
from ._definition import Definition
from ._search_view_type import SearchViews
from . import lookup, metadata



# coding: utf-8

from typing import TYPE_CHECKING

from ._session import Session
from .event_code import EventCode
from ._session_cxn_type import SessionCxnType
from ._session_type import SessionType
from ..._tools import create_repr, urljoin, cached_property, parse_url

if TYPE_CHECKING:
    from ._session_cxn_factory import PlatformConnection


class PlatformSession(Session):
    """
    This class is designed for handling the session to Refinitiv Data Platform (RDP)
    or Deployed Platform (TREP)
    - Refinitiv Data Platform are including handling an authentication and
        a token management (including refreshing token),
        also handling a real-time service discovery to get
        the service websocket endpoint and initialize the login for streaming
    - Deployed Platform is including the login for streaming
    """

    type = SessionType.PLATFORM

    def __init__(
        self,
        app_key=None,
        grant=None,
        signon_control=None,
        deployed_platform_host=None,
        deployed_platform_username=None,
        dacs_position=None,
        dacs_application_id=None,
        on_state=None,
        on_event=None,
        name="default",
        auto_reconnect=None,
        server_mode=None,
        base_url=None,
        auth_url=None,
        auth_authorize=None,
        auth_token=None,
        realtime_distribution_system_url=None,
    ):
        super().__init__(
            app_key,
            on_state=on_state,
            on_event=on_event,
            deployed_platform_username=deployed_platform_username,
            dacs_position=dacs_position,
            dacs_application_id=dacs_application_id,
            name=name,
        )

        self._grant = grant
        self._take_signon_control = signon_control if signon_control else True

        self._auto_reconnect = auto_reconnect
        self._server_mode = server_mode
        self._base_url = base_url
        self._auth_url = auth_url
        self._auth_authorize = auth_authorize
        self._auth_token = auth_token
        self._realtime_dist_system_url = realtime_distribution_system_url

        self._deployed_platform_host = deployed_platform_host
        self._deployed_platform_connection_name = self.name

        if self._deployed_platform_host is None and self._realtime_dist_system_url:
            parse_result = parse_url(self._realtime_dist_system_url)
            self._deployed_platform_host = parse_result.netloc
            self.debug(
                f"Using the Refinitiv realtime distribution system : "
                f"url at {self._realtime_dist_system_url},\n"
                f"deployed_platform_host={self._deployed_platform_host}"
            )

        elif self._deployed_platform_host and not self._realtime_dist_system_url:
            self.debug(
                f"Using the specific "
                f"deployed_platform_host={self._deployed_platform_host}"
            )

        elif self._deployed_platform_host and self._realtime_dist_system_url:
            # what to do ?
            pass

    @property
    def stream_auto_reconnection(self):
        return self._auto_reconnect

    @property
    def server_mode(self):
        return self._server_mode

    @property
    def signon_control(self):
        return self._take_signon_control

    @property
    def authentication_token_endpoint_url(self) -> str:
        url = urljoin(
            self._get_rdp_url_root() or "", self._auth_url or "", self._auth_token or ""
        )
        return url

    @cached_property
    def _connection(self) -> "PlatformConnection":
        from ._session_cxn_factory import get_session_cxn

        cxn_type = self._get_session_cxn_type()
        cxn = get_session_cxn(cxn_type, self)

        if cxn_type is SessionCxnType.REFINITIV_DATA:
            from .event import UpdateEvent

            auth_manager = cxn.auth_manager
            auth_manager.on(
                UpdateEvent.AUTHENTICATION_SUCCESS,
                lambda message: self._call_on_event(
                    EventCode.SessionAuthenticationSuccess, message
                ),
            )
            auth_manager.on(
                UpdateEvent.AUTHENTICATION_FAILED,
                lambda message: self._call_on_event(
                    EventCode.SessionAuthenticationFailed, message
                ),
            )
            auth_manager.on(
                UpdateEvent.RECONNECTING,
                lambda message: self._call_on_event(
                    EventCode.SessionReconnecting, "Session is reconnecting"
                ),
            )
            auth_manager.on(UpdateEvent.UPDATE_ACCESS_TOKEN, self.update_access_token)

        self.debug(f"Created session connection {cxn_type}")
        return cxn

    def _get_session_cxn_type(self) -> SessionCxnType:
        if self._grant and self._grant.is_valid() and self._deployed_platform_host:
            cxn_type = SessionCxnType.REFINITIV_DATA_AND_DEPLOYED

        elif self._grant and self._grant.is_valid():
            cxn_type = SessionCxnType.REFINITIV_DATA

        elif self._deployed_platform_host:
            cxn_type = SessionCxnType.DEPLOYED

        else:
            raise AttributeError(f"Can't get a session connection type")

        return cxn_type

    def _get_rdp_url_root(self):
        return self._base_url

    def _get_auth_token_uri(self):
        auth_token_uri = urljoin(self._auth_url, self._auth_token)
        uri = urljoin(self._get_rdp_url_root(), auth_token_uri)
        return uri

    def get_omm_login_message(self):
        return self._connection.get_omm_login_message()

    def get_rdp_login_message(self, stream_id):
        return {
            "streamID": f"{stream_id:d}",
            "method": "Auth",
            "token": self._access_token,
        }

    async def http_request_async(
        self,
        url: str,
        method=None,
        headers=None,
        data=None,
        params=None,
        json=None,
        closure=None,
        **kwargs,
    ):
        return await self._connection.http_request_async(
            url,
            method=method,
            headers=headers,
            data=data,
            params=params,
            json=json,
            closure=closure,
            **kwargs,
        )

    def http_request(
        self,
        url: str,
        method=None,
        headers=None,
        data=None,
        params=None,
        json=None,
        closure=None,
        **kwargs,
    ):
        return self._connection.http_request(
            url,
            method=method,
            headers=headers,
            data=data,
            params=params,
            json=json,
            closure=closure,
            **kwargs,
        )

    def __repr__(self):
        return create_repr(
            self,
            middle_path="session.platform",
            content=f"{{session_name='{self.name}'}}",
        )

    def __str__(self) -> str:
        s = (
            f"{self.__class__.__name__} {{\n"
            f"\t\tname={self.name},\n"
            f"\t\tserver_mode={self.server_mode},\n"
            f"\t\tstream_auto_reconnection={self.stream_auto_reconnection},\n"
            f"\t\tsignon_control={self.signon_control},\n"
            f"\t\tauthentication_token_endpoint_url={self.authentication_token_endpoint_url}}}"
        )
        return s



from ._enums import StatTypes, Frequency, SortOrder
from .._content_provider import UniverseContentValidator, ErrorParser
from .._join_responses import join_responses
from ..._tools import ArgsParser
from ..._tools import universe_arg_parser, make_enum_arg_parser
from ..._tools._datetime import ownership_datetime_adapter
from ...delivery._data._data_provider import (
    DataProvider,
    RequestFactory,
    ValidatorContainer,
)

MAX_LIMIT = 100


def get_correct_data_for_df(content_data):
    if not isinstance(content_data, list):
        return content_data

    if not content_data:
        return dict()

    _data = []
    for i in content_data:
        _data.extend(i["data"])

    headers = content_data[0].get("headers")
    if headers and isinstance(headers[0], list):
        headers = headers[0]

    result = {"data": _data, "headers": headers}
    return result


def parse_str(param):
    if isinstance(param, str):
        return param
    raise ValueError(f"Invalid type, expected str: {type(param)} is given")


class OwnershipRequestFactory(RequestFactory):
    def get_query_parameters(self, *_, **kwargs) -> list:
        query_parameters = []
        universe = kwargs.get("universe")
        if isinstance(universe, list):
            universe = list(set(universe))
        universe = universe_arg_parser.get_str(universe, delim=",")
        query_parameters.append(("universe", universe))

        stat_type = kwargs.get("stat_type")
        if stat_type is not None:
            stat_type = stat_types_ownership_arg_parser.get_str(stat_type)
            query_parameters.append(("statType", stat_type))

        offset = kwargs.get("offset")
        if offset is not None:
            query_parameters.append(("offset", offset))

        limit = kwargs.get("limit")
        if limit is not None:
            query_parameters.append(("limit", limit))

        sort_order = kwargs.get("sort_order")
        if sort_order is not None:
            sort_order = sort_order_ownership_arg_parser.get_str(sort_order)
            query_parameters.append(("sortOrder", sort_order))

        frequency = kwargs.get("frequency")
        if frequency is not None:
            frequency = frequency_ownership_arg_parser.get_str(frequency)
            query_parameters.append(("frequency", frequency))

        start = kwargs.get("start")
        if start is not None:
            start = ownership_datetime_adapter.get_str(start)
            query_parameters.append(("start", start))

        end = kwargs.get("end")
        if end is not None:
            end = ownership_datetime_adapter.get_str(end)
            query_parameters.append(("end", end))

        count = kwargs.get("count")
        if count is not None:
            query_parameters.append(("count", count))

        return query_parameters

    def extend_query_parameters(self, query_parameters, extended_params):
        # query_parameters -> [("param1", "val1"), ]
        result = dict(query_parameters)
        # result -> {"param1": "val1"}
        result.update(extended_params)
        # result -> {"param1": "val1", "extended_param": "value"}
        # return [("param1", "val1"), ("extended_param", "value")]
        return list(result.items())


class OwnershipDataProvider(DataProvider):
    def get_data(self, *args, **kwargs):
        limit = kwargs.get("limit")

        if limit is None:
            response = super().get_data(*args, **kwargs)

        else:
            responses = []

            for offset in range(0, limit, MAX_LIMIT):
                _kwargs = kwargs
                _kwargs["limit"] = (
                    MAX_LIMIT if offset < limit - MAX_LIMIT else limit - offset
                )
                _kwargs["offset"] = offset if offset else None
                response = super().get_data(*args, **kwargs)
                responses.append(response)

            response = join_responses(responses, reset_index=True)

        return response

    async def get_data_async(self, *args, **kwargs):
        limit = kwargs.get("limit")

        if limit is None:
            response = await super().get_data_async(*args, **kwargs)

        else:
            responses = []

            for offset in range(0, limit, MAX_LIMIT):
                _kwargs = kwargs
                _kwargs["limit"] = (
                    MAX_LIMIT if offset < limit - MAX_LIMIT else limit - offset
                )
                _kwargs["offset"] = offset if offset else None
                response = await super().get_data_async(*args, **kwargs)
                responses.append(response)

            response = join_responses(responses, reset_index=True)

        return response


universe_ownership_arg_parser = ArgsParser(parse_str)
stat_types_ownership_arg_parser = make_enum_arg_parser(StatTypes)
sort_order_ownership_arg_parser = make_enum_arg_parser(SortOrder)
frequency_ownership_arg_parser = make_enum_arg_parser(Frequency)

ownership_data_provider = OwnershipDataProvider(
    request=OwnershipRequestFactory(),
    validator=ValidatorContainer(content_validator=UniverseContentValidator()),
    parser=ErrorParser(),
)



class PackageNotFoundException(Exception):
    pass


class KeyNotFoundException(Exception):
    pass


class UpdateDBException(Exception):
    pass


class TableNotFoundException(Exception):
    pass


class DBModuleNotFoundException(Exception):
    pass


class ModuleNotFoundException(Exception):
    pass



__all__ = (
    "BondPricingError",
    "DesktopSessionError",
    "EndpointError",
    "EnvError",
    "ESGError",
    "FinancialContractsError",
    "NewsHeadlinesError",
    "PlatformSessionError",
    "RDError",
    "RequiredError",
    "SessionError",
    "StreamConnectionError",
    "StreamError",
    "StreamingError",
    "StreamingPricesError",
    "ConfigurationError",
)

from ._errors import (
    BondPricingError,
    DesktopSessionError,
    EndpointError,
    EnvError,
    ESGError,
    FinancialContractsError,
    NewsHeadlinesError,
    PlatformSessionError,
    RDError,
    RequiredError,
    SessionError,
    StreamConnectionError,
    StreamError,
    StreamingError,
    StreamingPricesError,
    ConfigurationError,
)



# coding: utf8
# contract_gen 2020-05-18 08:30:59.209815

__all__ = ["ProjectedIndexCalculationMethod"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class ProjectedIndexCalculationMethod(Enum):
    """
    Flag used to define how projected index is computed.

    Avalaible values are:
    - "ConstantIndex" : future index values are considered as constant and equal to
      projected index value.
    - "ForwardIndex" : future index values are computed using a forward curve.
    """

    CONSTANT_INDEX = "ConstantIndex"
    FORWARD_INDEX = "ForwardIndex"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(
            ProjectedIndexCalculationMethod,
            _PROJECTED_INDEX_CALCULATION_METHOD_VALUES,
            some,
        )

    @staticmethod
    def normalize(some):
        return _normalize(
            _PROJECTED_INDEX_CALCULATION_METHOD_VALUES_IN_LOWER_BY_PROJECTED_INDEX_CALCULATION_METHOD,
            some,
        )


_PROJECTED_INDEX_CALCULATION_METHOD_VALUES = tuple(
    t.value for t in ProjectedIndexCalculationMethod
)
_PROJECTED_INDEX_CALCULATION_METHOD_VALUES_IN_LOWER_BY_PROJECTED_INDEX_CALCULATION_METHOD = {
    name.lower(): item
    for name, item in ProjectedIndexCalculationMethod.__members__.items()
}



# coding: utf8

from enum import Enum, unique
from ..._enums._common_tools import _convert_to_str, _normalize


@unique
class MarketDataAccessDeniedFallback(Enum):
    IGNORE_CONSTITUENTS = "IgnoreConstituents"
    RETURN_ERROR = "ReturnError"
    USE_DELAYED_DATA = "UseDelayedData"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(
            MarketDataAccessDeniedFallback,
            _MARKET_DATA_ACCESS_DENIED_FALLBACK_VALUES,
            some,
        )

    @staticmethod
    def normalize(some):
        return _normalize(
            _MARKET_DATA_ACCESS_DENIED_FALLBACK_VALUES_IN_LOWER_BY_MARKET_DATA_ACCESS_DENIED_FALLBACK,
            some,
        )


_MARKET_DATA_ACCESS_DENIED_FALLBACK_VALUES = tuple(
    t.value for t in MarketDataAccessDeniedFallback
)
_MARKET_DATA_ACCESS_DENIED_FALLBACK_VALUES_IN_LOWER_BY_MARKET_DATA_ACCESS_DENIED_FALLBACK = {
    name.lower(): item
    for name, item in MarketDataAccessDeniedFallback.__members__.items()
}



# coding: utf8
# contract_gen 2020-05-18 08:30:59.250817

__all__ = ["SwaptionVolatilityType"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class SwaptionVolatilityType(Enum):
    ATM_SURFACE = "AtmSurface"
    SABR_VOLATILITY_CUBE = "SabrVolatilityCube"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(
            SwaptionVolatilityType, _SWAPTION_VOLATILITY_TYPE_VALUES, some
        )

    @staticmethod
    def normalize(some):
        return _normalize(
            _SWAPTION_VOLATILITY_TYPE_VALUES_IN_LOWER_BY_SWAPTION_VOLATILITY_TYPE, some
        )


_SWAPTION_VOLATILITY_TYPE_VALUES = tuple(t.value for t in SwaptionVolatilityType)
_SWAPTION_VOLATILITY_TYPE_VALUES_IN_LOWER_BY_SWAPTION_VOLATILITY_TYPE = {
    name.lower(): item for name, item in SwaptionVolatilityType.__members__.items()
}



from typing import TYPE_CHECKING, Optional, Iterable

if TYPE_CHECKING:
    from . import FxSurfaceParameters
    from ._models import VolatilitySurfacePoint
    from ._models._surface_output import SurfaceOutput
    from .._enums import Format


SurfaceParameters = Optional["FxSurfaceParameters"]
SurfaceLayout = Optional["SurfaceOutput"]
OptFormat = Optional["Format"]
OptVolatilitySurfacePoints = Optional[Iterable["VolatilitySurfacePoint"]]



# coding: utf8
# contract_gen 2020-05-18 08:30:59.211843

__all__ = ["Rounding"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class Rounding(Enum):
    DEFAULT = "Default"
    EIGHT = "Eight"
    FIVE = "Five"
    FOUR = "Four"
    ONE = "One"
    SEVEN = "Seven"
    SIX = "Six"
    THREE = "Three"
    TWO = "Two"
    UNROUNDED = "Unrounded"
    ZERO = "Zero"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(Rounding, _ROUNDING_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_ROUNDING_VALUES_IN_LOWER_BY_ROUNDING, some)


_ROUNDING_VALUES = tuple(t.value for t in Rounding)
_ROUNDING_VALUES_IN_LOWER_BY_ROUNDING = {
    name.lower(): item for name, item in Rounding.__members__.items()
}



# coding: utf8

__all__ = ["Frequency"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class Frequency(Enum):
    """
    The time frequency
    """

    EVERYDAY = "Everyday"
    BI_MONTHLY = "BiMonthly"
    MONTHLY = "Monthly"
    QUARTERLY = "Quarterly"
    SEMI_ANNUAL = "SemiAnnual"
    ANNUAL = "Annual"
    EVERY7_DAYS = "Every7Days"
    EVERY14_DAYS = "Every14Days"
    EVERY28_DAYS = "Every28Days"
    EVERY30_DAYS = "Every30Days"
    EVERY91_DAYS = "Every91Days"
    EVERY182_DAYS = "Every182Days"
    EVERY364_DAYS = "Every364Days"
    EVERY365_DAYS = "Every365Days"
    EVERY90_DAYS = "Every90Days"
    EVERY92_DAYS = "Every92Days"
    EVERY93_DAYS = "Every93Days"
    EVERY180_DAYS = "Every180Days"
    EVERY183_DAYS = "Every183Days"
    EVERY184_DAYS = "Every184Days"
    EVERY4_MONTHS = "Every4Months"
    EVERY_WORKING_DAY = "EveryWorkingDay"
    R2 = "R2"
    R4 = "R4"
    SCHEDULED = "Scheduled"
    ZERO = "Zero"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(Frequency, _FREQUENCY_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_FREQUENCY_VALUES_IN_LOWER_BY_FREQUENCY, some)


_FREQUENCY_VALUES = tuple(t.value for t in Frequency)
_FREQUENCY_VALUES_IN_LOWER_BY_FREQUENCY = {
    name.lower(): item for name, item in Frequency.__members__.items()
}



import asyncio
from functools import partial
from typing import TYPE_CHECKING

from pyee import EventEmitter

from .event import StreamStateEvent
from .stream_state import StreamState
from ..._core.log_reporter import _LogReporter
from ..._tools import cached_property

if TYPE_CHECKING:
    from logging import Logger


class StreamStateManager(_LogReporter):
    def __init__(self, logger: "Logger") -> None:
        _LogReporter.__init__(self, logger=logger)
        self._prev_state = None

        if not hasattr(self, "_classname"):
            self._classname: str = self.__class__.__name__

        self._state = StreamState.Closed

    @cached_property
    def _emitter(self) -> EventEmitter:
        return EventEmitter()

    @property
    def state(self) -> StreamState:
        return self._state

    def on(self, event, listener):
        return self._emitter.on(event, listener)

    def off(self, event, listener):
        return self._emitter.remove_listener(event, listener)

    @property
    def is_opened(self) -> bool:
        return self.state is StreamState.Opened

    @property
    def is_opening(self) -> bool:
        return self.state is StreamState.Opening

    @property
    def is_open(self) -> bool:
        return self.is_opened or self.is_opening

    @property
    def is_paused(self) -> bool:
        return self.state is StreamState.Paused

    @property
    def is_pausing(self) -> bool:
        return self.state is StreamState.Pausing

    @property
    def is_close(self) -> bool:
        return self.is_closed or self.is_closing

    @property
    def is_closing(self) -> bool:
        return self.state is StreamState.Closing

    @property
    def is_closed(self) -> bool:
        return self.state is StreamState.Closed

    def open(self, *args, **kwargs) -> StreamState:
        if self.is_opening or self.is_opened:
            self._warning(f"{self._classname} can’t open, state={self.state}")
            return self.state

        self._debug(f"{self._classname} is opening [o]")
        self._state = StreamState.Opening
        self._emitter.emit(StreamStateEvent.OPENING, self)
        self._do_open(*args, **kwargs)
        self._state = StreamState.Opened
        self._emitter.emit(StreamStateEvent.OPENED, self)
        self._debug(f"{self._classname} opened [O]")
        return self.state

    async def open_async(self, *args, **kwargs) -> StreamState:
        loop = asyncio.get_event_loop()
        result = await loop.run_in_executor(None, partial(self.open, *args, **kwargs))
        return result

    def _do_open(self, *args, **kwargs):
        # for override
        pass

    def close(self, *args, **kwargs) -> StreamState:
        if self.is_closing or self.is_closed:
            return self.state

        self._debug(f"{self._classname} is closing [c]")
        self._state = StreamState.Closing
        self._emitter.emit(StreamStateEvent.CLOSING, self)
        self._do_close(*args, **kwargs)
        self._state = StreamState.Closed
        self._emitter.emit(StreamStateEvent.CLOSED, self)
        self._debug(f"{self._classname} closed [C]")
        return self.state

    async def close_async(self, *args, **kwargs) -> StreamState:
        return self.close(*args, **kwargs)

    def _do_close(self, *args, **kwargs):
        # for override
        pass

    def halt(self, *args) -> StreamState:
        if self.is_closing or self.is_closed:
            return self.state

        self._debug(f"{self._classname} halt")
        self._dispose()
        return self.state

    def _dispose(self):
        # for override
        pass

    def pause(self):
        if self.is_paused or self.is_pausing:
            return self.state

        self._set_pause()
        self._do_pause()

        return self.state

    def _set_pause(self):
        self._prev_state = self.state
        self._state = StreamState.Paused

    def _do_pause(self):
        # for override
        pass

    def resume(self):
        if self.is_paused:
            self._set_resume()
            self._do_resume()
        return self.state

    def _set_resume(self):
        self._state = self._prev_state

    def _do_resume(self):
        # for override
        pass



from ..._ipa_content_provider import IPAContentProviderLayer
from ..._surfaces._eti_surface_request_item import EtiSurfaceRequestItem
from ...._content_type import ContentType


class Definition(IPAContentProviderLayer):
    def __init__(
        self,
        surface_layout,
        surface_parameters,
        underlying_definition,
        surface_tag,
        extended_params=None,
    ):
        request_item = EtiSurfaceRequestItem(
            surface_layout=surface_layout,
            surface_parameters=surface_parameters,
            underlying_definition=underlying_definition,
            surface_tag=surface_tag,
        )
        super().__init__(
            content_type=ContentType.SURFACES,
            universe=request_item,
            extended_params=extended_params,
        )



from abc import abstractmethod
from dataclasses import dataclass
from typing import Any


class BaseSpecification(object):
    @abstractmethod
    def is_satisfied_by(self, candidate: Any) -> bool:
        raise NotImplementedError()

    def and_(self, other: "BaseSpecification") -> "AndSpecification":
        return AndSpecification(self, other)

    def or_(self, other: "BaseSpecification") -> "OrSpecification":
        return OrSpecification(self, other)

    def not_(self) -> "NotSpecification":
        return NotSpecification(self)


@dataclass(frozen=True)
class AndSpecification(BaseSpecification):
    first: BaseSpecification
    second: BaseSpecification

    def is_satisfied_by(self, candidate: Any) -> bool:
        return self.first.is_satisfied_by(candidate) and self.second.is_satisfied_by(
            candidate
        )


@dataclass(frozen=True)
class OrSpecification(BaseSpecification):
    first: BaseSpecification
    second: BaseSpecification

    def is_satisfied_by(self, candidate: Any) -> bool:
        return self.first.is_satisfied_by(candidate) or self.second.is_satisfied_by(
            candidate
        )


@dataclass(frozen=True)
class NotSpecification(BaseSpecification):
    subject: BaseSpecification

    def is_satisfied_by(self, candidate: Any) -> bool:
        return not self.subject.is_satisfied_by(candidate)



# coding: utf-8
__version__ = "1.0.0b16"
__all__ = (
    "close_session",
    "content",
    "delivery",
    "eikon",
    "errors",
    "get_config",
    "get_data",
    "get_history",
    "load_config",
    "open_pricing_stream",
    "open_session",
    "OpenState",
    "PricingStream",
    "session",
)

"""
    refinitiv-data is a Python library to access Refinitiv Data Platform with Python.
"""

from ._config_functions import get_config, load_config
from ._fin_coder_layer.session import open_session, close_session
from ._fin_coder_layer import get_data, get_history
from ._fin_coder_layer.get_stream import PricingStream, open_pricing_stream
from ._open_state import OpenState
from . import session, delivery, content, errors, eikon



# coding: utf8

from enum import Enum, unique

from ..._enums._common_tools import _convert_to_str, _normalize


@unique
class CalendarAdjustment(Enum):
    FALSE = False
    NULL = None
    WEEKEND = "Weekend"
    CALENDAR = "Calendar"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(CalendarAdjustment, _CALENDAR_ADJUSTMENT_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(
            _CALENDAR_ADJUSTMENT_VALUES_IN_LOWER_BY_CALENDAR_ADJUSTMENT, some
        )


_CALENDAR_ADJUSTMENT_VALUES = tuple(t.value for t in CalendarAdjustment)
_CALENDAR_ADJUSTMENT_VALUES_IN_LOWER_BY_CALENDAR_ADJUSTMENT = {
    name.lower(): item for name, item in CalendarAdjustment.__members__.items()
}



from enum import Enum, unique


@unique
class SortOrder(Enum):
    old_to_new = "oldToNew"
    new_to_old = "newToOld"

    @staticmethod
    def convert_to_str(sort_order):
        if isinstance(sort_order, SortOrder):
            return sort_order.value
        elif sort_order in _SORT_ORDER_VALUES:
            return sort_order
        else:
            raise AttributeError(f"Sort order value must be in {_SORT_ORDER_VALUES}")


_SORT_ORDER_VALUES = [k.value for k in SortOrder]



# coding: utf8
# contract_gen 2020-06-16 10:26:10.694191

__all__ = ["InputVolatilityType"]

from enum import Enum, unique

from ._common_tools import _convert_to_str, _normalize


@unique
class InputVolatilityType(Enum):
    DEFAULT = "Default"
    LOG_NORMAL_VOLATILITY = "LogNormalVolatility"
    NORMALIZED_VOLATILITY = "NormalizedVolatility"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(InputVolatilityType, _INPUT_VOLATILITY_TYPE_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(
            _INPUT_VOLATILITY_TYPE_VALUES_IN_LOWER_BY_INPUT_VOLATILITY_TYPE, some
        )


_INPUT_VOLATILITY_TYPE_VALUES = tuple(t.value for t in InputVolatilityType)
_INPUT_VOLATILITY_TYPE_VALUES_IN_LOWER_BY_INPUT_VOLATILITY_TYPE = {
    name.lower(): item for name, item in InputVolatilityType.__members__.items()
}



# coding: utf8
# contract_gen 2020-05-13 12:48:48.744016

__all__ = ["Seniority"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class Seniority(Enum):
    SECURED = "Secured"
    SENIOR_UNSECURED = "SeniorUnsecured"
    SUBORDINATED = "Subordinated"
    JUNIOR_SUBORDINATED = "JuniorSubordinated"
    PREFERENCE = "Preference"
    NONE = "None"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(Seniority, _SENIORITY_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_SENIORITY_VALUES_IN_LOWER_BY_SENIORITY, some)


_SENIORITY_VALUES = tuple(t.value for t in Seniority)
_SENIORITY_VALUES_IN_LOWER_BY_SENIORITY = {
    name.lower(): item for name, item in Seniority.__members__.items()
}



# coding: utf8
import copy
import re
from enum import Enum, unique
from typing import Union, List, Tuple, TYPE_CHECKING

from .._content_provider import HistoricalDataProvider
from .._content_type import ContentType
from .._intervals import (
    DayIntervalType,
    Intervals,
    interval_arg_parser,
    get_day_interval_type,
)
from .._join_responses import join_historical_responses
from ..._tools import (
    urljoin,
    fields_arg_parser,
    make_enum_arg_parser,
)
from ..._tools._datetime import hp_datetime_adapter
from ...delivery._data._data_provider import (
    ResponseFactory,
    ContentValidator,
    RequestFactory,
)

if TYPE_CHECKING:
    from ...delivery._data._data_provider import Response

user_has_no_permissions_expr = re.compile(
    r"TS\.((Interday)|(Intraday))\.UserNotPermission\.[0-9]{5}"
)


# --------------------------------------------------------------------------------------
#   EventTypes
# --------------------------------------------------------------------------------------


@unique
class EventTypes(Enum):
    """
    The list of market events (comma delimiter), supported event types are trade,
    quote and correction.
    Note: Currently support only single event type.
        If request with multiple event types,
        the backend will pick up the first event type to proceed.
    """

    TRADE = "trade"
    QUOTE = "quote"
    CORRECTION = "correction"


event_types_arg_parser = make_enum_arg_parser(EventTypes)

# --------------------------------------------------------------------------------------
#   Intervals
# --------------------------------------------------------------------------------------


axis_by_day_interval_type = {
    DayIntervalType.INTRA: "Timestamp",
    DayIntervalType.INTER: "Date",
}

content_type_by_day_interval_type = {
    DayIntervalType.INTER: ContentType.HISTORICAL_PRICING_INTERDAY_SUMMARIES,
    DayIntervalType.INTRA: ContentType.HISTORICAL_PRICING_INTRADAY_SUMMARIES,
}


def get_content_type_by_interval(
    interval: Union[str, Intervals, DayIntervalType]
) -> ContentType:
    day_interval_type = get_day_interval_type(interval)
    return content_type_by_day_interval_type.get(day_interval_type)


# --------------------------------------------------------------------------------------
#   Adjustments
# --------------------------------------------------------------------------------------


@unique
class Adjustments(Enum):
    """
    The list of adjustment types (comma delimiter) that tells the system whether
     to apply or not apply CORAX (Corporate Actions) events or
     exchange/manual corrections to historical time series data.

     The supported values of adjustments :

        UNADJUSTED - Not apply both exchange/manual corrections and CORAX
        EXCHANGE_CORRECTION - Apply exchange correction adjustment to historical pricing
        MANUAL_CORRECTION - Apply manual correction adjustment to historical pricing
                            i.e. annotations made by content analysts
        CCH - Apply Capital Change adjustment to historical Pricing due
              to Corporate Actions e.g. stock split
        CRE - Apply Currency Redenomination adjustment
              when there is redenomination of currency
        RPO - Apply Reuters Price Only adjustment
              to adjust historical price only not volume
        RTS - Apply Reuters TimeSeries adjustment
              to adjust both historical price and volume
        QUALIFIERS - Apply price or volume adjustment
              to historical pricing according to trade/quote qualifier
              summarization actions
    """

    UNADJUSTED = "unadjusted"
    EXCHANGE_CORRECTION = "exchangeCorrection"
    MANUAL_CORRECTION = "manualCorrection"
    CCH = "CCH"
    CRE = "CRE"
    RPO = "RPO"
    RTS = "RTS"
    QUALIFIERS = "qualifiers"


adjustments_arg_parser = make_enum_arg_parser(Adjustments)


# --------------------------------------------------------------------------------------
#   MarketSession
# --------------------------------------------------------------------------------------


@unique
class MarketSession(Enum):
    """
    The marketsession parameter represents a list of interested official durations
        in which trade and quote activities occur for a particular universe.

    The supported values of marketsession :

        PRE - specifies that data returned
              should include data during pre-market session
        NORMAL - specifies that data returned
                 should include data during normal market session
        POST - specifies that data returned
               should include data during post-market session
    """

    PRE = "pre"
    NORMAL = "normal"
    POST = "post"


market_sessions_arg_parser = make_enum_arg_parser(MarketSession)


# --------------------------------------------------------------------------------------
#   ContentValidator
# --------------------------------------------------------------------------------------


class HistoricalPricingContentValidator(ContentValidator):
    def validate(self, data, *args, **kwargs):
        is_valid = True
        content_data = data.get("content_data")

        if not content_data:
            is_valid = False

        elif isinstance(content_data, list) and len(content_data):
            content_data = content_data[0]
            status = content_data.get("status", {})
            code = status.get("code", "")

            if status and user_has_no_permissions_expr.match(code):
                is_valid = False
                data["status"]["error"] = status
                data["errors"] = [(code, status.get("message"))]

            elif "Error" in code:
                is_valid = False
                data["status"]["error"] = status
                data["errors"] = [(code, status.get("message"))]

                if not (content_data.keys() - {"universe", "status"}):
                    is_valid = False

                elif "UserRequestError" in code:
                    is_valid = True

            elif not content_data.get("data"):
                is_valid = False

        return is_valid


# --------------------------------------------------------------------------------------
#   Response
# --------------------------------------------------------------------------------------


response_errors = {
    "default": "{error_message}. Requested ric: {rics}. Requested fields: {fields}",
    "TS.Intraday.UserRequestError.90001": "{rics} - The universe is not found",
    "TS.Intraday.Warning.95004": "{rics} - Trades interleaving with corrections is "
    "currently not supported. Corrections will not be returned.",
    "TS.Intraday.UserRequestError.90006": "{error_message} Requested ric: {rics}",
}


class HistoricalPricingResponseFactory(ResponseFactory):
    def get_raw(self, data):
        raw = data.get("content_data", [{}])
        raw = raw[0]
        return raw

    def create_success(self, data, *args, **kwargs):
        raw = self.get_raw(data)
        error_code = data.get("status_code") or (
            raw.get("status").get("code") if raw.get("status") else None
        )
        if error_code:
            self._compile_error_message(error_code, data, **kwargs)
        return super().create_success(data, *args, **kwargs)

    def create_fail(self, data, *args, **kwargs):
        raw = self.get_raw(data)
        status = raw.get("status", {})
        error_code = data.get("error_code", status.get("code"))
        self._compile_error_message(error_code, data, **kwargs)
        return super().create_fail(data, *args, **kwargs)

    @staticmethod
    def _compile_error_message(error_code: str, data: dict, **kwargs):
        """Compile error message in human readable format."""
        content_data = data.get("content_data")[0] if data.get("content_data") else {}
        error_message = data.get("error_message") or content_data.get("status", {}).get(
            "message"
        )
        fields = kwargs.get("fields")
        rics = (
            content_data.get("universe").get("ric")
            if content_data
            else kwargs.get("universe")
        )

        if error_code not in response_errors.keys():
            # Need to add error_code to data because different structure of responses
            data["error_code"] = error_code
            data["error_message"] = response_errors["default"].format(
                error_message=error_message, rics=rics, fields=fields
            )
        else:
            data["error_code"] = error_code
            data["error_message"] = response_errors[error_code].format(
                rics=rics, error_message=error_message
            )


# --------------------------------------------------------------------------------------
#   Request
# --------------------------------------------------------------------------------------


class HistoricalPricingRequestFactory(RequestFactory):
    def get_url(self, *args, **kwargs):
        url = args[1]
        url = urljoin(url, "/{universe}")
        return url

    def get_path_parameters(self, *_, **kwargs):
        universe = kwargs.get("universe")
        if universe is None:
            return {}
        return {"universe": universe}

    def get_field_timestamp(self, *args, **kwargs):
        return "DATE_TIME"

    def get_query_parameters(self, *args, **kwargs):
        query_parameters = []

        #
        # start
        #
        start = kwargs.get("start")
        if start:
            start = hp_datetime_adapter.get_str(start)
            query_parameters.append(("start", start))

        #
        # end
        #
        end = kwargs.get("end")
        if end:
            end = hp_datetime_adapter.get_str(end)
            query_parameters.append(("end", end))

        #
        # adjustments
        #
        adjustments = kwargs.get("adjustments")
        if adjustments:
            adjustments = adjustments_arg_parser.get_str(adjustments, delim=",")
            query_parameters.append(("adjustments", adjustments))

        #
        # market_sessions
        #
        market_sessions = kwargs.get("sessions")
        if market_sessions:
            market_sessions = market_sessions_arg_parser.get_str(
                market_sessions, delim=","
            )
            query_parameters.append(("sessions", market_sessions))

        #
        # count
        #
        count = kwargs.get("count") or 1
        if count < 1:
            raise AttributeError("Count minimum value is 1")

        if count > 1:
            query_parameters.append(("count", count))

        #
        # fields
        #
        fields = copy_fields(kwargs.get("fields"))
        if fields:
            fields = fields_arg_parser.get_list(fields)
            field_timestamp = self.get_field_timestamp(*args, **kwargs)
            if field_timestamp not in fields:
                fields.append(field_timestamp)
            query_parameters.append(("fields", ",".join(fields)))

        return query_parameters


class HistoricalPricingEventsRequestFactory(HistoricalPricingRequestFactory):
    def get_query_parameters(self, *args, **kwargs):
        query_parameters = super().get_query_parameters(*args, **kwargs)

        #
        # event_types
        #
        event_types = kwargs.get("event_types")
        if event_types:
            event_types = event_types_arg_parser.get_str(event_types, delim=",")
            query_parameters.append(("eventTypes", event_types))

        return query_parameters


field_timestamp_by_day_interval_type = {
    DayIntervalType.INTER: "DATE",
    DayIntervalType.INTRA: "DATE_TIME",
}


class HistoricalPricingSummariesRequestFactory(HistoricalPricingRequestFactory):
    def get_field_timestamp(self, *args, day_interval_type=None, **kwargs):
        return field_timestamp_by_day_interval_type.get(day_interval_type)

    def get_query_parameters(self, *args, **kwargs):
        query_parameters = super().get_query_parameters(*args, **kwargs)

        #
        # interval
        #
        interval = kwargs.get("interval")
        if interval:
            interval = interval_arg_parser.get_str(interval)
            query_parameters.append(("interval", interval))

        return query_parameters


# --------------------------------------------------------------------------------------
#   Providers
# --------------------------------------------------------------------------------------


class HPEventsDataProvider(HistoricalDataProvider):
    @staticmethod
    def _join_responses(
        responses: List[Tuple[str, "Response"]],
        fields: List[str],
        kwargs,
    ) -> "Response":
        axis_name = "Timestamp"
        return join_historical_responses(responses, axis_name, fields)


class HPSummariesDataProvider(HistoricalDataProvider):
    @staticmethod
    def _join_responses(
        responses: List[Tuple[str, "Response"]],
        fields: List[str],
        kwargs,
    ) -> "Response":
        axis_name = axis_by_day_interval_type.get(kwargs.get("day_interval_type"))
        return join_historical_responses(responses, axis_name, fields)


hp_events_data_provider = HPEventsDataProvider(
    request=HistoricalPricingEventsRequestFactory(),
    response=HistoricalPricingResponseFactory(),
    validator=HistoricalPricingContentValidator(),
)

hp_summaries_data_provider = HPSummariesDataProvider(
    request=HistoricalPricingSummariesRequestFactory(),
    response=HistoricalPricingResponseFactory(),
    validator=HistoricalPricingContentValidator(),
)


def copy_fields(fields: List[str]) -> List[str]:
    if fields is None:
        return []

    if not isinstance(fields, list):
        raise AttributeError(f"fields not support type {type(fields)}")

    return fields[:]



from typing import Union, Optional, TYPE_CHECKING

from .._ownership_data_provider import universe_ownership_arg_parser
from ..._content_provider import ContentProviderLayer
from ..._content_type import ContentType
from ...._tools import validate_types, validate_bool_value

if TYPE_CHECKING:
    from .._enums import Frequency
    from ....content._types import ExtendedParams

optional_date = Optional[Union[str, "datetime.datetime"]]


class Definition(ContentProviderLayer):
    """
    This class describe parameters to retrieve the report of the total consolidated shareholders
    invested in the requested company, at the specified historical period.

    Parameters
    ----------
    universe: str
        The Universe parameter allows the user to define the single company for which the content is returned.

    frequency: str, Frequency
        The frequency parameter allows users to request the frequency of the time series data, either quarterly or monthly.
        Available values : M, Q

    start: str, datetime, optional
        The start parameter allows users to define the start date of a time series.
        Dates are to be defined either by absolute or relative syntax.
        Example, "20190529", "-1Q", "1D", "-3MA", datetime.datetime.now().

    end: str, datetime, optional
        The end parameter allows users to define the start date of a time series.
        Dates are to be defined either by absolute or relative syntax.
        Example, "20190529", "-1Q", "1D", "-3MA", datetime.datetime.now().

    limit: int, optional
        The limit parameter is used for paging. It allows users to select the number of records to be returned.

    use_field_names_in_headers: bool, optional
        Return field name as column headers for data instead of title

    extended_params : ExtendedParams, optional
        If necessary other parameters.

    Examples
    --------
    >>> from refinitiv.data.content import ownership
    >>> definition = ownership.consolidated.shareholders_history_report.Definition("TRI.N", ownership.Frequency.MONTHLY)
    >>> response = definition.get_data()
    """

    def __init__(
        self,
        universe: str,
        frequency: Union[str, "Frequency"],
        start: optional_date = None,
        end: optional_date = None,
        limit: Optional[int] = None,
        use_field_names_in_headers: bool = False,
        extended_params: "ExtendedParams" = None,
    ):
        universe = universe_ownership_arg_parser.parse(universe)
        validate_types(limit, [int, type(None)], "limit")
        validate_bool_value(use_field_names_in_headers)

        super().__init__(
            content_type=ContentType.OWNERSHIP_CONSOLIDATED_SHAREHOLDERS_HISTORY_REPORT,
            universe=universe,
            frequency=frequency,
            start=start,
            end=end,
            limit=limit,
            use_field_names_in_headers=use_field_names_in_headers,
            extended_params=extended_params,
        )



from ._data_classes import Headline, Story, StoryUDF, HeadlineUDF
from ._sort_order import SortOrder
from ._tools import get_headlines_rdp
from .._join_responses import join_responses
from ..._core.session import DesktopSession
from ..._tools import to_datetime, make_enum_arg_parser
from ...delivery._data._data_provider import (
    DataProvider,
    RequestFactory,
    ResponseFactory,
    Response,
    Data,
    ContentValidator,
    ValidatorContainer,
)
from ...delivery._data._endpoint_data import RequestMethod

MAX_LIMIT = 100


# --------------------------------------------------------------------------------------
#   Content data
# --------------------------------------------------------------------------------------


# NewsStory


class NewsStoryData(Data):
    def __init__(self, raw, **kwargs):
        super().__init__(raw, **kwargs)
        self._story = None

    @property
    def story(self):
        if self._story is None:
            self._story = self.__class__._story_builder(self.raw)
        return self._story


class NewsStoryRDPData(NewsStoryData):
    _story_builder = Story.create


class NewsStoryUDFData(NewsStoryData):
    _story_builder = StoryUDF.create


# NewsHeadlines


class NewsHeadlinesData(Data):
    def __init__(self, raw, **kwargs):
        super().__init__(raw, **kwargs)
        self._headlines = None
        self._limit = None
        self._headlines_builder = None

    @property
    def headlines(self):
        if self._headlines is None:
            self._headlines = get_headlines_rdp(
                self.raw, self.__class__._headlines_builder, self._limit
            )

        return self._headlines


class NewsHeadlinesRDPData(NewsHeadlinesData):
    _headlines_builder = Headline.create


class NewsHeadlinesUDFData(NewsHeadlinesData):
    _headlines_builder = HeadlineUDF.create


# --------------------------------------------------------------------------------------
#   Response object
# --------------------------------------------------------------------------------------


class NewsStoryResponse(Response):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        _raw = None
        if self.is_success:
            _raw = self.data
        self._data = NewsStoryData(_raw)

    def __str__(self):
        if self._data and self._data.raw:
            return self._data.raw["newsItem"]["contentSet"]["inlineData"][0]["$"]
        else:
            return f"{self.errors}"

    @property
    def html(self):
        if self._data and self._data.raw:
            return (
                self._data.raw.get("newsItem", {})
                .get("contentSet", {})
                .get("inlineXML", [{}])[0]
                .get("$")
            )
        else:
            return None

    @property
    def text(self):
        if self._data and self._data.raw:
            return self._data.raw["newsItem"]["contentSet"]["inlineData"][0]["$"]
        else:
            return None


# --------------------------------------------------------------------------------------
#   Content data validator
# --------------------------------------------------------------------------------------


class NewsUDFContentValidator(ContentValidator):
    def validate(self, data, *args, **kwargs):
        is_valid = super().validate(data)
        content_data = data.get("content_data")
        if not is_valid:
            return is_valid

        if "ErrorCode" in content_data:
            data["error_code"] = content_data.get("ErrorCode")
            data["error_message"] = content_data.get("ErrorMessage")
            is_valid = False
        elif not content_data:
            is_valid = False
            data["error_code"] = 1
            data["error_message"] = "Content data not contain any data"

        return is_valid


# --------------------------------------------------------------------------------------
#   Response factory
# --------------------------------------------------------------------------------------


class NewsStoryResponseFactory(ResponseFactory):
    @staticmethod
    def change_code_and_message_in_data(data: dict) -> dict:
        new_error_msg = "Error while calling the NEP backend: Story not found"
        error_code = data.get("error_code") or 0
        error_msg = data.get("error_message") or ""

        if error_code == 400 or error_code == 404 and new_error_msg != error_msg:
            data["error_code"] = 404
            data["error_message"] = new_error_msg

        return data

    def create_fail(self, *args, **kwargs):
        data = args[0]
        self.change_code_and_message_in_data(data)
        return super().create_fail(*args, **kwargs)


# --------------------------------------------------------------------------------------
#   Request factory
# --------------------------------------------------------------------------------------

# NewsStory


class NewsStoryUDFRequestFactory(RequestFactory):
    def get_body_parameters(self, *args, **kwargs):
        entity = {
            "E": "News_Story",
        }
        w = dict()

        story_id = kwargs.get("story_id")
        w["storyId"] = story_id

        session = args[0]
        app_key = session.app_key
        w["productName"] = app_key

        entity["W"] = w
        body_parameters = {"Entity": entity}
        return body_parameters

    def get_url(self, *args, **kwargs):
        session = args[0]
        url = session._get_rdp_url_root()
        if isinstance(session, DesktopSession):
            url = session._get_udf_url()
        return url

    def update_url(self, url_root, url, path_parameters, query_parameters):
        return url

    def get_request_method(self, *_, **kwargs) -> RequestMethod:
        return RequestMethod.POST


class NewsStoryRDPRequestFactory(RequestFactory):
    def get_path_parameters(self, *_, **kwargs):
        story_id = kwargs.get("story_id")
        path_parameters = {"storyId": story_id}
        return path_parameters

    def get_header_parameters(self, *args, **kwargs):
        header_parameters = {"accept": "application/json"}
        return header_parameters

    def get_url(self, *args, **kwargs):
        return super().get_url(*args, **kwargs) + "/{storyId}"


# NewsHeadlines


class NewsHeadlinesUDFRequestFactory(RequestFactory):
    def extend_body_parameters(self, body_parameters, extended_params):
        body_parameters["Entity"]["W"].update(extended_params)
        return body_parameters

    def get_body_parameters(self, *args, **kwargs):
        entity = {
            "E": "News_Headlines",
        }
        w = dict()

        query = kwargs.get("query")
        w["query"] = query

        count = kwargs.get("count")
        if count is not None:
            w["number"] = str(count)

        payload = kwargs.get("payload")
        if payload is not None:
            w["payload"] = payload

        repository = kwargs.get("repository")
        if repository is not None:
            w["repository"] = repository

        session = args[0]
        app_key = session.app_key
        w["productName"] = app_key

        date_from = kwargs.get("date_from")
        if date_from is not None:
            w["dateFrom"] = to_datetime(date_from).isoformat()

        date_to = kwargs.get("date_to")
        if date_to is not None:
            w["dateTo"] = to_datetime(date_to).isoformat()

        entity["W"] = w
        body_parameters = {"Entity": entity}
        return body_parameters

    def get_url(self, *args, **kwargs):
        session = args[0]
        url = session._get_rdp_url_root()
        if isinstance(session, DesktopSession):
            url = session._get_udf_url()
        return url

    def update_url(self, url_root, url, path_parameters, query_parameters):
        return url

    def get_request_method(self, *_, **kwargs) -> RequestMethod:
        return RequestMethod.POST


class NewsHeadlinesRDPRequestFactory(RequestFactory):
    def extend_query_parameters(self, query_parameters, extended_params):
        # query_parameters -> [("param1", "val1"), ]
        result = dict(query_parameters)
        # result -> {"param1": "val1"}
        result.update(extended_params)
        # result -> {"param1": "val1", "extended_param": "value"}
        # return [("param1", "val1"), ("extended_param", "value")]
        return list(result.items())

    def get_query_parameters(self, *_, **kwargs):
        query_parameters = []

        query = kwargs.get("query")
        query_parameters.append(("query", query))

        count = kwargs.get("count")
        if count is not None:
            query_parameters.append(("limit", count))

        date_from = kwargs.get("date_from")
        if date_from is not None:
            date_from = to_datetime(date_from).isoformat()
            query_parameters.append(("dateFrom", date_from))

        date_to = kwargs.get("date_to")
        if date_to is not None:
            date_to = to_datetime(date_to).isoformat()
            query_parameters.append(("dateTo", date_to))

        sort_order = kwargs.get("sort_order")
        if sort_order is not None:
            sort_order = sort_order_news_arg_parser.get_str(sort_order)
            query_parameters.append(("sort", sort_order))

        # for pagination
        cursor = kwargs.get("cursor")
        if cursor is not None:
            query_parameters.append(("cursor", cursor))

        return query_parameters


# --------------------------------------------------------------------------------------
#   Data provider
# --------------------------------------------------------------------------------------

# NewsStory


class NewsUDFDataProvider(DataProvider):
    def get_data(self, *args, **kwargs):
        limit = kwargs.get("count")

        if limit and limit > MAX_LIMIT:
            kwargs["count"] = MAX_LIMIT

            responses = []
            for _ in range(0, limit, MAX_LIMIT):
                response = super().get_data(*args, **kwargs)
                responses.append(response)

            response = join_responses(responses, data_class=self.response.data_class)
            response.data._limit = limit

        else:
            response = super().get_data(*args, **kwargs)

        return response

    async def get_data_async(self, *args, **kwargs):
        limit = kwargs.get("count")

        if limit and limit > MAX_LIMIT:
            kwargs["count"] = MAX_LIMIT

            responses = []
            for _ in range(0, limit, MAX_LIMIT):
                response = await super().get_data_async(*args, **kwargs)
                responses.append(response)

            response = join_responses(responses, data_class=self.response.data_class)
            response.data._limit = limit

        else:
            response = await super().get_data_async(*args, **kwargs)

        return response


# NewsHeadlines


class NewsHeadlinesRDPDataProvider(DataProvider):
    def get_data(self, *args, **kwargs):
        on_page_response = kwargs.get("on_page_response")
        limit = kwargs.get("count")
        responses = []

        if limit > MAX_LIMIT:
            kwargs["count"] = MAX_LIMIT

        for _ in range(0, limit, MAX_LIMIT):
            response = super().get_data(*args, **kwargs)
            responses.append(response)

            if on_page_response:
                on_page_response(self, response)

            meta = response.data.raw.get("meta", {})
            kwargs = {
                "cursor": meta.get("next"),
                "__content_type__": kwargs.get("__content_type__"),
            }

        response = join_responses(responses, data_class=NewsHeadlinesRDPData)
        response.data._limit = limit
        return response

    async def get_data_async(self, *args, **kwargs):
        on_page_response = kwargs.get("on_page_response")
        limit = kwargs.get("count")
        responses = []

        if limit > MAX_LIMIT:
            kwargs["count"] = MAX_LIMIT

        for _ in range(0, limit, MAX_LIMIT):
            response = await super().get_data_async(*args, **kwargs)
            responses.append(response)

            if on_page_response:
                on_page_response(self, response)

            meta = response.data.raw.get("meta", {})
            kwargs = {
                "cursor": meta.get("next"),
                "__content_type__": kwargs.get("__content_type__"),
            }

        response = join_responses(responses, data_class=NewsHeadlinesRDPData)
        response.data._limit = limit
        return response


sort_order_news_arg_parser = make_enum_arg_parser(SortOrder)

validator = ValidatorContainer(content_validator=NewsUDFContentValidator())

# NewsStory

news_story_data_provider_rdp = DataProvider(
    response=NewsStoryResponseFactory(
        response_class=NewsStoryResponse,
        data_class=NewsStoryRDPData,
    ),
    request=NewsStoryRDPRequestFactory(),
)

news_story_data_provider_udf = NewsUDFDataProvider(
    response=NewsStoryResponseFactory(
        response_class=NewsStoryResponse,
        data_class=NewsStoryUDFData,
    ),
    request=NewsStoryUDFRequestFactory(),
    validator=validator,
)

# NewsHeadlines

news_headlines_data_provider_rdp = NewsHeadlinesRDPDataProvider(
    response=ResponseFactory(data_class=NewsHeadlinesRDPData),
    request=NewsHeadlinesRDPRequestFactory(),
)
news_headlines_data_provider_udf = NewsUDFDataProvider(
    response=ResponseFactory(data_class=NewsHeadlinesUDFData),
    request=NewsHeadlinesUDFRequestFactory(),
    validator=validator,
)



# coding: utf8

__all__ = ["CdsConvention"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class CdsConvention(Enum):
    ISDA = "ISDA"
    USER_DEFINED = "UserDefined"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(CdsConvention, _CDS_CONVENTION_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_CDS_CONVENTION_VALUES_IN_LOWER_BY_CDS_CONVENTION, some)


_CDS_CONVENTION_VALUES = tuple(t.value for t in CdsConvention)
_CDS_CONVENTION_VALUES_IN_LOWER_BY_CDS_CONVENTION = {
    name.lower(): item for name, item in CdsConvention.__members__.items()
}



# coding: utf8

__all__ = ["UnderlyingType"]

from enum import Enum, unique

from ._common_tools import _convert_to_str, _normalize


@unique
class UnderlyingType(Enum):
    CAP = "Cap"
    ETI = "Eti"
    FX = "Fx"
    SWAPTION = "Swaption"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(UnderlyingType, _UNDERLYING_TYPE_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_UNDERLYING_TYPE_VALUES_IN_LOWER_BY_UNDERLYING_TYPE, some)


_UNDERLYING_TYPE_VALUES = tuple(t.value for t in UnderlyingType)
_UNDERLYING_TYPE_VALUES_IN_LOWER_BY_UNDERLYING_TYPE = {
    name.lower(): item for name, item in UnderlyingType.__members__.items()
}



import os
from typing import Optional, TYPE_CHECKING, Union

import httpx

from . import SessionType
from ._retry_transport import RequestRetryException, RetryAsyncTransport, RetryTransport
from ..log_reporter import LogReporter
from ... import _configure as configure
from ..._tools import cached_property

APPLICATION_JSON = "application/json"

if TYPE_CHECKING:
    from ._session import Session


def get_http_limits(config):
    max_connections = config.get(configure.keys.http_max_connections)
    max_keepalive_connections = config.get(
        configure.keys.http_max_keepalive_connections
    )
    limits = httpx.Limits(
        max_connections=max_connections,
        max_keepalive_connections=max_keepalive_connections,
    )
    return limits


def get_http_request_timeout_secs(session):
    """the default http request timeout in secs"""
    key = configure.keys.http_request_timeout
    value = session.config.get(key)

    is_list = isinstance(value, list)
    if is_list and len(value) == 1:
        value = value[0]
        try:
            value = int(value)
        except ValueError:
            pass

    number = isinstance(value, int) or isinstance(value, float)
    negative_number = number and value < 0

    if number and value == 0:
        value = None
    elif number and value == 1:
        value = 1

    is_none = value is None

    set_default = not is_none and (not number or negative_number)

    if set_default:
        value = configure.defaults.http_request_timeout
        session.warning(f"Invalid value of the {key}. Default value is used")

    return value


class BaseHTTPClient(LogReporter):
    def __init__(self, session: "Session") -> None:
        super().__init__(session)
        self._session: "Session" = session

        self._client: Optional[httpx.Client, httpx.AsyncClient] = None
        self._auto_retry_client: Optional[httpx.Client, httpx.AsyncClient] = None

    def is_closed(self) -> bool:
        is_closed = self._client is None
        return is_closed or self._client.is_closed

    def _filtered_timeout(self, kwargs):
        http_request_timeout = None
        if httpx.__version__ < "0.20.0":
            http_request_timeout = kwargs.pop(
                "timeout", get_http_request_timeout_secs(self._session)
            )
        else:
            if "timeout" not in kwargs:
                #   override using the http request timeout from config file
                kwargs["timeout"] = get_http_request_timeout_secs(self._session)
        return http_request_timeout

    def build_request(
        self,
        client: Union[httpx.AsyncClient, httpx.Client],
        url: str,
        method=None,
        headers=None,
        data=None,
        params=None,
        json=None,
        closure=None,
        **kwargs,
    ) -> httpx.Request:
        if method is None:
            method = "GET"

        if headers is None:
            headers = {}

        access_token = self._session._access_token
        app_key = self._session.app_key

        if access_token is not None:
            headers["Authorization"] = "Bearer {}".format(access_token)

        if closure is not None:
            headers["Closure"] = closure

        cookies = None

        if self._session.type == SessionType.DESKTOP:
            proxy_app_version = os.getenv("DP_PROXY_APP_VERSION")
            user_uuid = os.getenv("REFINITIV_AAA_USER_ID")

            if proxy_app_version:
                headers.update({"app-version": proxy_app_version})

            if user_uuid:
                cookies = {"user_uuid": user_uuid}

        headers.update({"x-tr-applicationid": app_key})

        if httpx.__version__ < "0.20.0":
            kwargs.pop("timeout", None)

        self.debug(
            f"Request to {url}\n"
            f"\tmethod = {method}\n"
            f"\theaders = {headers}\n"
            f"\tparams = {params}\n"
            f"\tcookies = {cookies}\n"
            f"\tdata = {data}\n"
            f"\tjson = {json}"
        )
        request = client.build_request(
            method=method,
            url=url,
            data=data,
            json=json,
            params=params,
            headers=headers,
            cookies=cookies,
            **kwargs,
        )
        return request


class HTTPClient(BaseHTTPClient):
    def open(self):
        config = self._session.config
        limits = get_http_limits(config)

        # httpx has its default Accept header and
        # server wants application/json or nothing
        self._client = httpx.Client(headers={"Accept": APPLICATION_JSON}, limits=limits)

        key = configure.keys.http_auto_retry_config
        auto_retry_config = config.get(key, None)

        if auto_retry_config:
            number_of_retries = auto_retry_config.get("number-of-retries", 3)
            retry_on_errors = auto_retry_config.get("on-errors", [])
            retry_backoff_factor = auto_retry_config.get("backoff-factor", 1)
            retry_on_methods = auto_retry_config.get("on-methods", ["GET", "POST"])

            retry_transport = RetryTransport(
                total_attempts=number_of_retries,
                on_statuses=retry_on_errors,
                on_methods=retry_on_methods,
                backoff_factor=retry_backoff_factor,
            )
            self._auto_retry_client = httpx.Client(transport=retry_transport)

    def close(self):
        if self._client:
            self._client.close()
            self._client = None

        if self._auto_retry_client:
            self._auto_retry_client.close()
            self._auto_retry_client = None

    def send(
        self,
        url: str,
        method=None,
        headers=None,
        data=None,
        params=None,
        json=None,
        closure=None,
        **kwargs,
    ) -> httpx.Response:
        auto_retry = kwargs.pop("auto_retry", False)
        client: httpx.Client = self._auto_retry_client if auto_retry else self._client

        http_request_timeout = self._filtered_timeout(kwargs=kwargs)

        request = self.build_request(
            client=client,
            url=url,
            method=method,
            headers=headers,
            data=data,
            params=params,
            json=json,
            closure=closure,
            **kwargs,
        )

        if httpx.__version__ < "0.20.0":
            response = client.send(request, timeout=http_request_timeout)
        else:
            response = client.send(request)

        return response


class AsyncHTTPClient(BaseHTTPClient):
    async def open_async(self):
        config = self._session.config
        limits = get_http_limits(config)

        # httpx has its default Accept header and
        # server wants application/json or nothing
        self._client = httpx.AsyncClient(
            headers={"Accept": APPLICATION_JSON}, limits=limits
        )

        key = configure.keys.http_auto_retry_config
        auto_retry_config = config.get(key, None)

        if auto_retry_config:
            number_of_retries = auto_retry_config.get("number-of-retries", 3)
            retry_on_errors = auto_retry_config.get("on-errors", [])
            retry_backoff_factor = auto_retry_config.get("backoff-factor", 1)
            retry_on_methods = auto_retry_config.get("on-methods", ["GET", "POST"])

            retry_transport = RetryAsyncTransport(
                total_attempts=number_of_retries,
                on_statuses=retry_on_errors,
                on_methods=retry_on_methods,
                backoff_factor=retry_backoff_factor,
            )
            self._auto_retry_client = httpx.AsyncClient(transport=retry_transport)

    async def close_async(self):
        if self._client:
            self._client = httpx.AsyncClient(headers={"Accept": APPLICATION_JSON})
            await self._client.aclose()
            self._client = None

        if self._auto_retry_client:
            await self._auto_retry_client.aclose()
            self._auto_retry_client = None

    async def send(
        self,
        url: str,
        method=None,
        headers=None,
        data=None,
        params=None,
        json=None,
        closure=None,
        **kwargs,
    ) -> httpx.Response:
        auto_retry = kwargs.pop("auto_retry", False)
        client: httpx.AsyncClient = (
            self._auto_retry_client if auto_retry else self._client
        )

        http_request_timeout = self._filtered_timeout(kwargs=kwargs)

        request = self.build_request(
            client=client,
            url=url,
            method=method,
            headers=headers,
            data=data,
            params=params,
            json=json,
            closure=closure,
            **kwargs,
        )
        if httpx.__version__ < "0.20.0":
            response = await client.send(request, timeout=http_request_timeout)
        else:
            response = await client.send(request)

        return response


class HTTPService(LogReporter):
    def __init__(self, session: "Session") -> None:
        super().__init__(session)
        self._session: "Session" = session

    @cached_property
    def _client(self) -> HTTPClient:
        return HTTPClient(self._session)

    @cached_property
    def _client_async(self) -> AsyncHTTPClient:
        return AsyncHTTPClient(self._session)

    @property
    def request_timeout_secs(self):
        return get_http_request_timeout_secs(self._session)

    def open(self):
        self._client.open()

    def close(self):
        self._client.close()

    def request(
        self,
        url: str,
        method=None,
        headers=None,
        data=None,
        params=None,
        json=None,
        closure=None,
        **kwargs,
    ) -> httpx.Response:
        if self._client.is_closed():
            self._client.open()

        try:
            response = self._client.send(
                url=url,
                method=method,
                headers=headers,
                data=data,
                params=params,
                json=json,
                closure=closure,
                **kwargs,
            )
        except RequestRetryException as error:
            self.error(error)
            raise error
        except (
            httpx.ConnectError,
            httpx.ConnectTimeout,
            httpx.HTTPStatusError,
            httpx.InvalidURL,
            httpx.LocalProtocolError,
            httpx.NetworkError,
            httpx.ProtocolError,
            httpx.ProxyError,
            httpx.ReadError,
            httpx.RequestError,
            httpx.ReadTimeout,
            httpx.RemoteProtocolError,
            httpx.TooManyRedirects,
            httpx.TransportError,
            httpx.TimeoutException,
        ) as error:
            self.error(
                f"An error occurred while requesting {error.request.url!r}.\n"
                f"\t{error!r}"
            )
            raise error

        self.debug(f"HTTP request response {response.status_code}: {response.text}")
        return response

    async def open_async(self):
        await self._client_async.open_async()

    async def close_async(self):
        await self._client_async.close_async()

    async def request_async(
        self,
        url: str,
        method=None,
        headers=None,
        data=None,
        params=None,
        json=None,
        closure=None,
        **kwargs,
    ) -> httpx.Response:
        if self._client_async.is_closed():
            await self._client_async.open_async()

        try:
            response = await self._client_async.send(
                url=url,
                method=method,
                headers=headers,
                data=data,
                params=params,
                json=json,
                closure=closure,
                **kwargs,
            )
        except RequestRetryException as error:
            self.error(error)
            raise error
        except (
            httpx.ConnectError,
            httpx.ConnectTimeout,
            httpx.HTTPStatusError,
            httpx.InvalidURL,
            httpx.LocalProtocolError,
            httpx.NetworkError,
            httpx.ProtocolError,
            httpx.ProxyError,
            httpx.ReadError,
            httpx.RequestError,
            httpx.ReadTimeout,
            httpx.RemoteProtocolError,
            httpx.TooManyRedirects,
            httpx.TransportError,
            httpx.TimeoutException,
        ) as error:
            self.error(
                f"An error occurred while requesting {error.request.url!r}.\n"
                f"\t{error!r}"
            )
            raise error

        self.debug(f"HTTP request response {response.status_code}: {response.text}")
        return response


def get_service(session: "Session") -> HTTPService:
    return HTTPService(session)



from typing import Optional, List

from .._content_type import ContentType

from ...delivery._data._data_provider import DataProviderLayer, Response
from ...delivery._data._endpoint_data import RequestMethod


class CustomInstrument:
    def __init__(
        self,
        symbol,
        formula,
        instrument_name=None,
        exchange_name=None,
        currency=None,
        time_zone=None,
        holidays=None,
        description=None,
        session=None,
    ):
        """
        symbol : str
            Instrument symbol in the format "S)someSymbol.YOURUUID".
        formula : str
            Formula consisting of rics (fields can be specified by comma).
        instrument_name : str, optional
            Human-readable name of the instrument. Maximum of 16 characters.
        exchange_name : str, optional
            4-letter code of the listing exchange.
        currency : str, optional
            3-letter code of the currency of the instrument, e.g. GBP.
        time_zone: str, optional
            Time Series uses an odd custom 3-letter value for time zone IDs, e.g. "LON" for London.
        holidays : list[dict], optional
            List of custom calendar definitions.
        description : str, optional
            Free text field from the user to put any notes or text. Up to 1000 characters.

        """
        self.symbol = symbol
        self.formula = formula
        self.instrument_name = instrument_name
        self.exchange_name = exchange_name
        self.currency = currency
        self.time_zone = time_zone
        self.holidays = holidays
        self.description = description
        self.id = None
        self.owner = None
        self.type = None
        self._session = session

    def delete(self):
        data_provider_layer = DataProviderLayer(
            data_type=ContentType.CUSTOM_INSTRUMENTS_INSTRUMENTS, universe=self.symbol
        )
        data_provider_layer.get_data(self._session, method=RequestMethod.DELETE)

    def _update(self, data):
        self.symbol = data.get("symbol")
        self.formula = data.get("formula")
        self.instrument_name = data.get("instrumentName")
        self.exchange_name = data.get("exchangeName")
        self.currency = data.get("currency")
        self.time_zone = data.get("timeZone")
        self.holidays = data.get("holidays")
        self.description = data.get("description")
        self.id = data.get("id")
        self.type = data.get("type")
        self.owner = data.get("owner")

    def save(self):
        data = {
            "symbol": self.symbol,
            "formula": self.formula,
            "instrument_name": self.instrument_name,
            "exchange_name": self.exchange_name,
            "currency": self.currency,
            "time_zone": self.time_zone,
            "holidays": self.holidays,
            "description": self.description,
            "id": self.id,
            "type": self.type,
            "owner": self.owner,
        }
        data_provider_layer = DataProviderLayer(
            data_type=ContentType.CUSTOM_INSTRUMENTS_INSTRUMENTS,
            universe=self.id,
            **data,
        )
        response = data_provider_layer.get_data(self._session, method=RequestMethod.PUT)

        self._update(response.data.raw)


def delete(
    universe,
    session=None,
) -> Response:
    """
    universe : str
        Instrument symbol in the format "S)someSymbol.YOURUUID".

    Examples
    --------
    >>> from refinitiv.data.content.custom_instruments.manage import delete
    >>> response = delete("MyInstrument")
    """
    data_provider_layer = DataProviderLayer(
        data_type=ContentType.CUSTOM_INSTRUMENTS_INSTRUMENTS, universe=universe
    )
    return data_provider_layer.get_data(session, method=RequestMethod.DELETE)


def get(universe, session=None) -> CustomInstrument:
    data_provider_layer = DataProviderLayer(
        data_type=ContentType.CUSTOM_INSTRUMENTS_INSTRUMENTS, universe=universe
    )
    response = data_provider_layer.get_data(method=RequestMethod.GET, session=session)
    data = response.data.raw
    symbol = data.get("symbol")
    formula = data.get("formula")
    ci = CustomInstrument(symbol=symbol, formula=formula)
    ci._update(data)
    return ci


def create(
    symbol: str,
    formula: str,
    instrument_name: Optional[str] = None,
    exchange_name: Optional[str] = None,
    currency: Optional[str] = None,
    time_zone: Optional[str] = None,
    holidays: Optional[List[dict]] = None,
    description: Optional[str] = None,
    session=None,
    on_response=None,
) -> CustomInstrument:
    """
    symbol : str
        Instrument symbol in the format "S)someSymbol.YOURUUID".
    formula : str
        Formula consisting of rics (fields can be specified by comma).
    instrument_name : str, optional
        Human-readable name of the instrument. Maximum of 16 characters.
    exchange_name : str, optional
        4-letter code of the listing exchange.
    currency : str, optional
        3-letter code of the currency of the instrument, e.g. GBP.
    time_zone: str, optional
        Time Series uses an odd custom 3-letter value for time zone IDs, e.g. "LON" for London.
    holidays : list[dict], optional
        List of custom calendar definitions.
    description : str, optional
        Free text field from the user to put any notes or text. Up to 1000 characters.

    Examples
    --------
    >>> from refinitiv.data.content.custom_instruments.manage import create
    >>> response = create(symbol="MyNewInstrument", formula="EUR=*3")
    """
    data_provider_layer = DataProviderLayer(
        data_type=ContentType.CUSTOM_INSTRUMENTS_INSTRUMENTS,
        symbol=symbol,
        formula=formula,
        instrument_name=instrument_name,
        exchange_name=exchange_name,
        currency=currency,
        time_zone=time_zone,
        holidays=holidays,
        description=description,
    )
    response = data_provider_layer.get_data(
        session, on_response, method=RequestMethod.POST
    )
    data = response.data.raw
    symbol = data.get("symbol")
    formula = data.get("formula")
    ci = CustomInstrument(symbol=symbol, formula=formula)
    ci._update(data)
    return ci



# coding: utf8
# contract_gen 2020-05-18 08:30:59.238824

__all__ = ["ForwardComputeMethod"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class ForwardComputeMethod(Enum):
    USE_DIVIDENDS_ONLY = "UseDividendsOnly"
    USE_FORWARD_CRV = "UseForwardCrv"
    USE_FORWARD_CRV_AND_DIVIDENDS = "UseForwardCrvAndDividends"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(
            ForwardComputeMethod, _FORWARD_COMPUTE_METHOD_VALUES, some
        )

    @staticmethod
    def normalize(some):
        return _normalize(
            _FORWARD_COMPUTE_METHOD_VALUES_IN_LOWER_BY_FORWARD_COMPUTE_METHOD, some
        )


_FORWARD_COMPUTE_METHOD_VALUES = tuple(t.value for t in ForwardComputeMethod)
_FORWARD_COMPUTE_METHOD_VALUES_IN_LOWER_BY_FORWARD_COMPUTE_METHOD = {
    name.lower(): item for name, item in ForwardComputeMethod.__members__.items()
}



import itertools
from typing import Dict, Type, Optional
from typing import TYPE_CHECKING
from typing import Union

from . import OMMStreamConnection, RDPStreamConnection
from ._omm_stream import _OMMStream
from ._protocol_type import ProtocolType
from ._rdp_stream import _RDPStream
from ._stream_cxn_config import StreamCxnConfig, get_cxn_config
from .._data._api_type import APIType
from ... import _log as log
from ...content._content_type import ContentType
from ...content._types import OptDict, OptStr, Strings, ExtendedParams, OptCall

if TYPE_CHECKING:
    from . import StreamConnection
    from ..._core.session import Session


def logger():
    return log.root_logger().getChild("stream-factory")


protocol_type_by_name: Dict[str, ProtocolType] = {
    "OMM": ProtocolType.OMM,
    "RDP": ProtocolType.RDP,
}

api_config_key_by_api_type: Dict[APIType, str] = {
    APIType.STREAMING_FINANCIAL_CONTRACTS: "streaming/quantitative-analytics/financial-contracts",
    APIType.STREAMING_PRICING: "streaming/pricing/main",
    APIType.STREAMING_TRADING: "streaming/trading-analytics/redi",
    APIType.STREAMING_BENCHMARK: "streaming/benchmark/resource",
    APIType.STREAMING_CUSTOM_INSTRUMENTS: "streaming/custom-instruments/resource",
}

api_type_by_content_type: Dict[ContentType, APIType] = {
    ContentType.STREAMING_CHAINS: APIType.STREAMING_PRICING,
    ContentType.STREAMING_PRICING: APIType.STREAMING_PRICING,
    ContentType.STREAMING_TRADING: APIType.STREAMING_TRADING,
    ContentType.STREAMING_CONTRACTS: APIType.STREAMING_FINANCIAL_CONTRACTS,
    ContentType.STREAMING_CUSTOM_INSTRUMENTS: APIType.STREAMING_CUSTOM_INSTRUMENTS,
}

connection_id_iterator = itertools.count(0)

stream_class_by_protocol_type: Dict[
    ProtocolType, Type[Union[_OMMStream, _RDPStream]]
] = {
    ProtocolType.OMM: _OMMStream,
    ProtocolType.RDP: _RDPStream,
}


class StreamDetails:
    def __init__(
        self,
        content_type: ContentType,
        protocol_type: ProtocolType,
        api_type: APIType,
        api_config_key: str = "",
    ) -> None:
        self.api_config_key = api_config_key
        self.api_type = api_type
        self.protocol_type = protocol_type
        self.content_type = content_type

        if self.api_type is ContentType.STREAMING_CUSTOM:
            self.api_type_as_str = f"{self.api_type}.{self.api_config_key}"
        else:
            self.api_type_as_str = str(self.api_type)


def content_type_to_details(content_type: ContentType) -> StreamDetails:
    api_type = api_type_by_content_type.get(content_type, APIType.STREAMING_CUSTOM)

    if content_type is ContentType.STREAMING_CUSTOM:
        raise ValueError("Cannot create StreamDetails, without api.")

    return StreamDetails(content_type, ProtocolType.NONE, api_type)


def convert_api_config_key_to_content_type(api_config_key: str) -> ContentType:
    """
    >>> api_type_by_api_config_key
    {
        'streaming/quantitative-analytics/financial-contracts': <APIType.STREAMING_FINANCIAL_CONTRACTS: 3>,
        'streaming/pricing/main': <APIType.STREAMING_PRICING: 8>,
        'streaming/trading-analytics/redi': <APIType.STREAMING_TRADING: 11>
    }
    >>> content_type_by_api_type
    {
        <APIType.STREAMING_PRICING: 8>: <ContentType.STREAMING_PRICING: 17>,
        <APIType.STREAMING_TRADING: 11>: <ContentType.STREAMING_TRADING: 39>,
        <APIType.STREAMING_FINANCIAL_CONTRACTS: 3>: <ContentType.STREAMING_CONTRACTS: 6>
    }
    """
    api_type_by_api_config_key = {v: k for k, v in api_config_key_by_api_type.items()}
    api_type = api_type_by_api_config_key.get(api_config_key)
    content_type_by_api_type = {v: k for k, v in api_type_by_content_type.items()}
    content_type = content_type_by_api_type.get(api_type)

    if not content_type:
        content_type = ContentType.STREAMING_CUSTOM

    return content_type


def get_valid_content_type(content_type: ContentType, api: str = "") -> ContentType:
    if content_type in {ContentType.STREAMING_CUSTOM, ContentType.NONE}:
        if not api:
            raise ValueError("api cannot be None")

        content_type = convert_api_config_key_to_content_type(api)

    return content_type


def create_omm_stream(
    content_type: ContentType,
    session: "Session",
    name: str,
    api: str = "",
    domain: OptStr = None,
    service: OptStr = None,
    fields: Optional[Strings] = None,
    key: OptDict = None,
    extended_params: "ExtendedParams" = None,
    on_refresh: OptCall = None,
    on_status: OptCall = None,
    on_update: OptCall = None,
    on_complete: OptCall = None,
    on_error: OptCall = None,
) -> _OMMStream:
    if content_type is ContentType.NONE and not api:
        content_type = ContentType.STREAMING_PRICING

    else:
        content_type = get_valid_content_type(content_type, api)

    api_type = api_type_by_content_type.get(content_type, APIType.STREAMING_CUSTOM)
    details = StreamDetails(content_type, ProtocolType.OMM, api_type, api)
    stream_id = next(session._omm_stream_counter)
    stream = _OMMStream(
        stream_id=stream_id,
        session=session,
        name=name,
        domain=domain,
        service=service,
        fields=fields,
        key=key,
        extended_params=extended_params,
        on_refresh=on_refresh,
        on_status=on_status,
        on_update=on_update,
        on_complete=on_complete,
        on_error=on_error,
        details=details,
    )
    logger().debug(f" + Created stream={stream.classname}")
    return stream


def create_rdp_stream(
    content_type: ContentType,
    session: "Session",
    universe: Union[list, dict],
    view: list,
    extended_params: "ExtendedParams",
    parameters: OptDict = None,
    service: OptStr = None,
    api: str = "",
    on_ack: OptCall = None,
    on_response: OptCall = None,
    on_update: OptCall = None,
    on_alarm: OptCall = None,
) -> _RDPStream:
    content_type = get_valid_content_type(content_type, api)
    api_type = api_type_by_content_type.get(content_type, APIType.STREAMING_CUSTOM)
    details = StreamDetails(content_type, ProtocolType.RDP, api_type, api)
    stream_id = next(session._rdp_stream_counter)
    stream = _RDPStream(
        stream_id=stream_id,
        session=session,
        service=service,
        universe=universe,
        view=view,
        parameters=parameters,
        extended_params=extended_params,
        on_ack=on_ack,
        on_response=on_response,
        on_update=on_update,
        on_alarm=on_alarm,
        details=details,
    )
    logger().debug(f" + Created stream={stream.classname}")
    return stream


def get_protocol_type_by_name(protocol_name: str) -> ProtocolType:
    protocol_type = protocol_type_by_name.get(protocol_name)

    if not protocol_type:
        raise ValueError(f"Can't find protocol type by name: {protocol_name}")

    return protocol_type


cxn_class_by_protocol_type: Dict[
    ProtocolType,
    Type[Union[OMMStreamConnection, RDPStreamConnection]],
] = {
    ProtocolType.OMM: OMMStreamConnection,
    ProtocolType.RDP: RDPStreamConnection,
}

_max_reconnect_default = 5


def load_config(details: StreamDetails, session: "Session") -> "StreamCxnConfig":
    content_type = details.content_type

    if content_type is ContentType.STREAMING_CUSTOM:
        api_config_key = details.api_config_key

        if not api_config_key:
            raise ValueError(
                "For ContentType.STREAMING_CUSTOM, api_config_key cannot be None"
            )

        api_type = api_config_key

    else:
        api_type = api_type_by_content_type.get(content_type)
        api_config_key = api_config_key_by_api_type.get(api_type)

    config: "StreamCxnConfig" = get_cxn_config(api_config_key, session)
    logger().debug(f"Loaded config for {api_type}, {config}")
    return config


def create_stream_cxn(details: StreamDetails, session: "Session") -> "StreamConnection":
    content_type = details.content_type
    protocol_type = details.protocol_type
    config = load_config(details, session)
    session_id = session.session_id
    connection_id = next(connection_id_iterator)
    name = f"{protocol_type.name}{content_type.name}_{session_id}.{connection_id}"
    cxn_class = cxn_class_by_protocol_type.get(protocol_type)
    cxn = cxn_class(
        connection_id=connection_id,
        name=name,
        session=session,
        config=config,
        max_reconnect=_max_reconnect_default,
    )
    logger().debug(
        f" + Created: \n"
        f"\tcxn={cxn},\n"
        f"\tconfig={config},\n"
        f"\tmax_reconnect={_max_reconnect_default}"
    )
    return cxn



# coding: utf8

from typing import Optional, Iterable

from ._term_deposit_definition import TermDepositInstrumentDefinition
from ._term_deposit_pricing_parameters import PricingParameters
from .._base_definition import BaseDefinition
from ..._enums import BusinessDayConvention, DateRollingConvention, DayCountBasis
from ....._tools import create_repr


class Definition(BaseDefinition):
    """
    API endpoint for Financial Contract analytics,
    that returns calculations relevant to each contract type.

    Parameters
    ----------
    instrument_code : str, optional
        Code to define the term deposit instrument. For the moment, only RICs for CDs
        and Wholesales deposits are supported, with deposit code (ex:"EUR1MD=").
    instrument_tag : str, optional
        User defined string to identify the instrument. It can be used to link output
        results to the instrument definition. Only alphabetic, numeric and '- _.#=@'
        characters are supported.
    start_date : str, optional
        The date the term deposit starts accruing interest. Its effective date.
        By default it is derived from the ValuationDate and the day to spot convention
        of the contract currency.
    end_date : str, optional
        The maturity date of the term deposit contract. Either the endDate or the tenor
        must be provided.
    tenor : str, optional
        The period code that represents the time between the start date and end date of
        the contract.
        Mandatory if instrumentCode is null. Either the endDate or the tenor must be
        provided.
    notional_ccy : str, optional
        The ISO code of the notional currency.
        Should be explicitly specified if InstrumentCode hasn't been specified.
        May be retrieved from reference data.
    notional_amount : float, optional
        The notional amount of the term deposit at the start date.
        By default 1,000,000 is used.
    fixed_rate_percent : float, optional
        Fixed interest rate percent to be applied for notional by deal terms.
        Mandatory if instrument_code is None.
    payment_business_day_convention : BusinessDayConvention, optional
        The method to adjust dates to a working day.
        By default 'ModifiedFollowing'.
    payment_roll_convention : DateRollingConvention, optional
        Method to adjust payment dates when they fall at the end of the month.
        By default 'Last'.
    year_basis : DayCountBasis, optional
        The Day Count Basis method used to calculate the interest payments.
        By default 'Dcb_Actual_365'.
    calendar : str, optional
        Calendar used to adjust deposit duration calculation.
        By default the calendar corresponding to notional currency is used.
    fields : list of str, optional
        Contains the list of Analytics that the quantitative analytic service will
        compute.
    pricing_parameters : PricingParameters, optional
        The pricing parameters to apply to this instrument. If pricing parameters
        are not provided at this level parameters defined globally at the request
        level are used. If no pricing parameters are provided globally default
        values apply.
    extended_params : dict, optional
        If necessary other parameters.

    Methods
    -------
    get_data(session=session, on_response=on_response, async_mode=None)
        Returns a response to the data platform
    get_data_async(session=session, on_response=on_response, async_mode=None)
        Returns a response to the async data platform
    get_stream(session=session, api="")
        Get stream quantitative analytic service subscription

    Examples
    --------
     >>> import refinitiv.data.content.ipa.financial_contracts as rdf
     >>> definition = rdf.term_deposit.Definition(
     ...   tenor="5Y",
     ...   notional_ccy="EUR",
     ...   fixed_rate_percent=11
     ...)
     >>> response = definition.get_data()

     Using get_stream
     >>> stream = definition.get_stream()
     >>> stream.open()
    """

    def __init__(
        self,
        instrument_code: Optional[str] = None,
        instrument_tag: Optional[str] = None,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
        tenor: Optional[str] = None,
        notional_ccy: Optional[str] = None,
        notional_amount: Optional[float] = None,
        fixed_rate_percent: Optional[float] = None,
        payment_business_day_convention: Optional[BusinessDayConvention] = None,
        payment_roll_convention: Optional[DateRollingConvention] = None,
        year_basis: Optional[DayCountBasis] = None,
        calendar: Optional[str] = None,
        fields: Optional[Iterable[str]] = None,
        pricing_parameters: Optional[PricingParameters] = None,
        extended_params: Optional[dict] = None,
    ):
        definition = TermDepositInstrumentDefinition(
            instrument_code=instrument_code,
            instrument_tag=instrument_tag,
            start_date=start_date,
            end_date=end_date,
            tenor=tenor,
            notional_ccy=notional_ccy,
            notional_amount=notional_amount,
            fixed_rate_percent=fixed_rate_percent,
            payment_business_day_convention=payment_business_day_convention,
            payment_roll_convention=payment_roll_convention,
            year_basis=year_basis,
            calendar=calendar,
        )
        super().__init__(
            definition=definition,
            fields=fields,
            pricing_parameters=pricing_parameters,
            extended_params=extended_params,
        )

    def __repr__(self):
        return create_repr(
            self,
            middle_path="content.ipa.financial_contracts.term_deposit",
        )



# coding: utf8

from enum import Enum, unique
from ...._enums._common_tools import (
    _convert_to_str,
    _normalize,
)


@unique
class SettlementType(Enum):
    ASSET = "Asset"
    CASH = "Cash"
    UNDEFINED = "Undefined"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(SettlementType, _SETTLEMENT_TYPE_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_SETTLEMENT_TYPE_VALUES_IN_LOWER_BY_SETTLEMENT_TYPE, some)


_SETTLEMENT_TYPE_VALUES = tuple(t.value for t in SettlementType)
_SETTLEMENT_TYPE_VALUES_IN_LOWER_BY_SETTLEMENT_TYPE = {
    name.lower(): item for name, item in SettlementType.__members__.items()
}



from typing import Union, Any

import numpy as np
import pandas as pd
from numpy import iterable

from ._enums import Axis
from ...delivery._data._data_provider import RequestFactory
from ...delivery.endpoint_request import RequestMethod
from ..._tools._common import ArgsParser

types_by_axis = {
    Axis.DATE: "datetime64",
    "startDate": "datetime64",
    "endDate": "datetime64",
    Axis.DELTA: float,
    Axis.EXPIRY: float,
    Axis.MONEYNESS: float,
    Axis.STRIKE: float,
    Axis.TENOR: float,
    "discountFactor": float,
    "ratePercent": float,
}


def get_type_by_axis(axis):
    axis_values_type = types_by_axis.get(axis)

    if not axis_values_type:
        raise ValueError(f"Cannot find axis's values type for axis {axis}.")

    return axis_values_type


def parse_value(value: Any) -> Union[float, int, np.datetime64]:
    if isinstance(value, str):
        try:
            value = int(value)
        except ValueError:
            pass
        else:
            return value

        try:
            value = pd.to_datetime(value).to_numpy()
        except ValueError:
            try:
                value = pd.to_datetime(value, dayfirst=True).to_numpy()
            except ValueError:
                pass
        else:
            return value

        try:
            value = float(value)
        except ValueError:
            raise ValueError(f"not valid format: {value}")

    return value


value_arg_parser = ArgsParser(parse_value)
x_arg_parser = value_arg_parser
y_arg_parser = value_arg_parser


# ---------------------------------------------------------------------------
#   RequestFactory
# ---------------------------------------------------------------------------


def parse_universe(universe):
    retval = []

    if not iterable(universe):
        universe = [universe]

    # convert universe's objects into json
    for i, item in enumerate(universe):
        if not hasattr(item, "get_json"):
            item = item._kwargs.get("universe")

        retval.append(item.get_json())

    return retval


def parse_outputs(outputs):
    retval = []

    if not isinstance(outputs, list):
        outputs = [outputs]

    for item in outputs:
        if hasattr(item, "value"):
            item = item.value

        retval.append(item)

    return retval


universe_arg_parser = ArgsParser(parse_universe)
outputs_arg_parser = ArgsParser(parse_outputs)


class CurvesAndSurfacesRequestFactory(RequestFactory):
    def get_request_method(self, *_, **kwargs):
        return kwargs.get("method", RequestMethod.POST)

    def get_body_parameters(self, *_, **kwargs):
        universe = kwargs.get("universe") or []
        outputs = kwargs.get("outputs")

        body_parameters = {}

        # universe
        universe = universe_arg_parser.get_list(universe)
        body_parameters["universe"] = universe

        # outputs
        if outputs:
            outputs = outputs_arg_parser.get_list(outputs)
            body_parameters["outputs"] = outputs

        return body_parameters



# coding: utf8

__all__ = ["BinaryType"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class BinaryType(Enum):
    DIGITAL = "Digital"
    NO_TOUCH = "NoTouch"
    NONE = "None"
    ONE_TOUCH = "OneTouch"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(BinaryType, _BINARY_TYPE_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_BINARY_TYPE_VALUES_IN_LOWER_BY_BINARY_TYPE, some)


_BINARY_TYPE_VALUES = tuple(t.value for t in BinaryType)
_BINARY_TYPE_VALUES_IN_LOWER_BY_BINARY_TYPE = {
    name.lower(): item for name, item in BinaryType.__members__.items()
}



# coding: utf8

from typing import Optional

from ._enums import (
    PriceSide,
    FxSwapCalculationMethod,
)
from ._models import FxPoint
from .._instrument_pricing_parameters import InstrumentPricingParameters


class PricingParameters(InstrumentPricingParameters):
    """
    API endpoint for Financial Contract analytics,
    that returns calculations relevant to each contract type.

    Parameters
    ----------
    deposit_ccy1 : FxPoint, optional

    deposit_ccy2 : FxPoint, optional

    fx_swap_calculation_method : FxSwapCalculationMethod, optional
        The method we chose to price outrights using or not implied deposits. Possible
        values are:   FxSwap (compute outrights using swap points),
        deposit_ccy1ImpliedFromFxSwap (compute currency1 deposits using swap points),
        deposit_ccy2ImpliedFromFxSwap (compute currency2 deposits using swap points).
        Optional. Defaults to 'FxSwap'.
    fx_swaps_ccy1 : FxPoint, optional

    fx_swaps_ccy1_ccy2 : FxPoint, optional

    fx_swaps_ccy2 : FxPoint, optional

    fx_swaps_far_leg : FxPoint, optional

    fx_swaps_near_leg : FxPoint, optional

    price_side : PriceSide, optional
        The type of price returned for pricing Analysis: Bid(Bid value), Ask(Ask value),
        Mid(Mid value) Optional. Defaults to 'Mid'.
    adjust_all_swap_points_to_cross_calendars : bool, optional
        This flag define if cross-calendar holidays should be taken into account for
        fx_swaps_ccy1 and fx_swaps_ccy2. it adjusts swap points according to cross
        holidays if it's set to true, by default set to false.
    calc_end_from_fwd_start : bool, optional

    calc_end_from_pre_spot_start : bool, optional

    ignore_ref_ccy_holidays : bool, optional
        The reference currency holidays flag : When dates are computed, its possible to
        choose if holidays of the reference currency are included or not in the pricing
        Optional. Defaults to 'false'.
    market_data_date : str, optional
        The market data date for pricing. Optional. By default, the market_data_date
        date is the valuation_date or Today
    one_day_values : str, optional

    roll_over_time_policy : str, optional

    spread_margin_in_bp : float, optional
        If activated, it will calculate the indicated points in market data section
        instead of taking them directly from the curves
    valuation_date : str, optional
        The valuation date for pricing.  Optional. If not set the valuation date is
        equal to market_data_date or Today. For assets that contains a
        settlementConvention, the default valuation date  is equal to the settlementdate
        of the Asset that is usually the TradeDate+SettlementConvention.
    """

    def __init__(
        self,
        deposit_ccy1: Optional[FxPoint] = None,
        deposit_ccy2: Optional[FxPoint] = None,
        fx_swap_calculation_method: Optional[FxSwapCalculationMethod] = None,
        fx_swaps_ccy1: Optional[FxPoint] = None,
        fx_swaps_ccy1_ccy2: Optional[FxPoint] = None,
        fx_swaps_ccy2: Optional[FxPoint] = None,
        fx_swaps_far_leg: Optional[FxPoint] = None,
        fx_swaps_near_leg: Optional[FxPoint] = None,
        price_side: Optional[PriceSide] = None,
        adjust_all_swap_points_to_cross_calendars: Optional[bool] = None,
        calc_end_from_fwd_start: Optional[bool] = None,
        calc_end_from_pre_spot_start: Optional[bool] = None,
        ignore_ref_ccy_holidays: Optional[bool] = None,
        market_data_date: Optional[str] = None,
        one_day_values: Optional[str] = None,
        roll_over_time_policy: Optional[str] = None,
        spread_margin_in_bp: Optional[float] = None,
        valuation_date: Optional[str] = None,
    ) -> None:
        super().__init__()
        self.deposit_ccy1 = deposit_ccy1
        self.deposit_ccy2 = deposit_ccy2
        self.fx_swap_calculation_method = fx_swap_calculation_method
        self.fx_swaps_ccy1 = fx_swaps_ccy1
        self.fx_swaps_ccy1_ccy2 = fx_swaps_ccy1_ccy2
        self.fx_swaps_ccy2 = fx_swaps_ccy2
        self.fx_swaps_far_leg = fx_swaps_far_leg
        self.fx_swaps_near_leg = fx_swaps_near_leg
        self.price_side = price_side
        self.adjust_all_swap_points_to_cross_calendars = (
            adjust_all_swap_points_to_cross_calendars
        )
        self.calc_end_from_fwd_start = calc_end_from_fwd_start
        self.calc_end_from_pre_spot_start = calc_end_from_pre_spot_start
        self.ignore_ref_ccy_holidays = ignore_ref_ccy_holidays
        self.market_data_date = market_data_date
        self.one_day_values = one_day_values
        self.roll_over_time_policy = roll_over_time_policy
        self.spread_margin_in_bp = spread_margin_in_bp
        self.valuation_date = valuation_date

    @property
    def deposit_ccy1(self):
        """
        :return: object FxPoint
        """
        return self._get_object_parameter(FxPoint, "depositCcy1")

    @deposit_ccy1.setter
    def deposit_ccy1(self, value):
        self._set_object_parameter(FxPoint, "depositCcy1", value)

    @property
    def deposit_ccy2(self):
        """
        :return: object FxPoint
        """
        return self._get_object_parameter(FxPoint, "depositCcy2")

    @deposit_ccy2.setter
    def deposit_ccy2(self, value):
        self._set_object_parameter(FxPoint, "depositCcy2", value)

    @property
    def fx_swap_calculation_method(self):
        """
        The method we chose to price outright using or not implied deposits. Possible
        values are:   FxSwap (compute outright using swap points),
        DepositCcy1ImpliedFromFxSwap (compute currency1 deposits using swap points),
        DepositCcy2ImpliedFromFxSwap (compute currency2 deposits using swap points).
        Optional. Defaults to 'FxSwap'.
        :return: enum FxSwapCalculationMethod
        """
        return self._get_enum_parameter(
            FxSwapCalculationMethod, "fxSwapCalculationMethod"
        )

    @fx_swap_calculation_method.setter
    def fx_swap_calculation_method(self, value):
        self._set_enum_parameter(
            FxSwapCalculationMethod, "fxSwapCalculationMethod", value
        )

    @property
    def fx_swaps_ccy1(self):
        """
        :return: object FxPoint
        """
        return self._get_object_parameter(FxPoint, "fxSwapsCcy1")

    @fx_swaps_ccy1.setter
    def fx_swaps_ccy1(self, value):
        self._set_object_parameter(FxPoint, "fxSwapsCcy1", value)

    @property
    def fx_swaps_ccy1_ccy2(self):
        """
        :return: object FxPoint
        """
        return self._get_object_parameter(FxPoint, "fxSwapsCcy1Ccy2")

    @fx_swaps_ccy1_ccy2.setter
    def fx_swaps_ccy1_ccy2(self, value):
        self._set_object_parameter(FxPoint, "fxSwapsCcy1Ccy2", value)

    @property
    def fx_swaps_ccy2(self):
        """
        :return: object FxPoint
        """
        return self._get_object_parameter(FxPoint, "fxSwapsCcy2")

    @fx_swaps_ccy2.setter
    def fx_swaps_ccy2(self, value):
        self._set_object_parameter(FxPoint, "fxSwapsCcy2", value)

    @property
    def fx_swaps_far_leg(self):
        """
        :return: object FxPoint
        """
        return self._get_object_parameter(FxPoint, "fxSwapsFarLeg")

    @fx_swaps_far_leg.setter
    def fx_swaps_far_leg(self, value):
        self._set_object_parameter(FxPoint, "fxSwapsFarLeg", value)

    @property
    def fx_swaps_near_leg(self):
        """
        :return: object FxPoint
        """
        return self._get_object_parameter(FxPoint, "fxSwapsNearLeg")

    @fx_swaps_near_leg.setter
    def fx_swaps_near_leg(self, value):
        self._set_object_parameter(FxPoint, "fxSwapsNearLeg", value)

    @property
    def price_side(self):
        """
        The type of price returned for pricing Analysis:
        Bid(Bid value),
        Ask(Ask value),
        Mid(Mid value)
        Optional. Defaults to 'Mid'.
        :return: enum PriceSide
        """
        return self._get_enum_parameter(PriceSide, "priceSide")

    @price_side.setter
    def price_side(self, value):
        self._set_enum_parameter(PriceSide, "priceSide", value)

    @property
    def adjust_all_swap_points_to_cross_calendars(self):
        """
        This flag define if cross-calendar holidays should be taken into account for
        fx_swaps_ccy1 and fx_swaps_ccy2. it adjusts swap points according to cross
        holidays if it's set to true, by default set to false.
        :return: bool
        """
        return self._get_parameter("adjustAllSwapPointsToCrossCalendars")

    @adjust_all_swap_points_to_cross_calendars.setter
    def adjust_all_swap_points_to_cross_calendars(self, value):
        self._set_parameter("adjustAllSwapPointsToCrossCalendars", value)

    @property
    def calc_end_from_fwd_start(self):
        """
        :return: bool
        """
        return self._get_parameter("calcEndFromFwdStart")

    @calc_end_from_fwd_start.setter
    def calc_end_from_fwd_start(self, value):
        self._set_parameter("calcEndFromFwdStart", value)

    @property
    def calc_end_from_pre_spot_start(self):
        """
        :return: bool
        """
        return self._get_parameter("calcEndFromPreSpotStart")

    @calc_end_from_pre_spot_start.setter
    def calc_end_from_pre_spot_start(self, value):
        self._set_parameter("calcEndFromPreSpotStart", value)

    @property
    def ignore_ref_ccy_holidays(self):
        """
        The reference currency holidays flag : When dates are computed, its possible to
        choose if holidays of the reference currency are included or not in the pricing
        Optional. Defaults to 'false'.
        :return: bool
        """
        return self._get_parameter("ignoreRefCcyHolidays")

    @ignore_ref_ccy_holidays.setter
    def ignore_ref_ccy_holidays(self, value):
        self._set_parameter("ignoreRefCcyHolidays", value)

    @property
    def market_data_date(self):
        """
        The market data date for pricing. Optional. By default, the market_data_date
        date is the valuation_date or Today
        :return: str
        """
        return self._get_parameter("marketDataDate")

    @market_data_date.setter
    def market_data_date(self, value):
        self._set_parameter("marketDataDate", value)

    @property
    def one_day_values(self):
        """
        :return: str
        """
        return self._get_parameter("oneDayValues")

    @one_day_values.setter
    def one_day_values(self, value):
        self._set_parameter("oneDayValues", value)

    @property
    def roll_over_time_policy(self):
        """
        :return: str
        """
        return self._get_parameter("rollOverTimePolicy")

    @roll_over_time_policy.setter
    def roll_over_time_policy(self, value):
        self._set_parameter("rollOverTimePolicy", value)

    @property
    def spread_margin_in_bp(self):
        """
        If activated, it will calculate the indicated points in market data section
        instead of taking them directly from the curves
        :return: float
        """
        return self._get_parameter("spreadMarginInBp")

    @spread_margin_in_bp.setter
    def spread_margin_in_bp(self, value):
        self._set_parameter("spreadMarginInBp", value)

    @property
    def valuation_date(self):
        """
        The valuation date for pricing.  Optional. If not set the valuation date is
        equal to market_data_date or Today. For assets that contains a
        settlementConvention, the default valuation date  is equal to the settlementdate
        of the Asset that is usually the TradeDate+SettlementConvention.
        :return: str
        """
        return self._get_parameter("valuationDate")

    @valuation_date.setter
    def valuation_date(self, value):
        self._set_parameter("valuationDate", value)



import re
from json import JSONDecodeError
from typing import Tuple, List

import pandas as pd
import requests

from .._content_provider import (
    HistoricalDataProvider,
    field_timestamp_by_day_interval_type,
    axis_by_day_interval_type,
)
from .._content_type import ContentType
from .._intervals import DayIntervalType, get_day_interval_type, Intervals
from .._join_responses import join_historical_responses
from ..._errors import RDError
from ..._tools import (
    get_response_reason,
    make_enum_arg_parser,
    custom_insts_datetime_adapter,
)
from ..._tools._dataframe import convert_df_columns_to_datetime
from ...delivery._data._data_provider import (
    RequestFactory,
    ResponseFactory,
    Parser,
    success_http_codes,
    ContentValidator,
    DataProvider,
    DataProviderLayer,
    ContentTypeValidator,
    ValidatorContainer,
    Response,
)
from ...delivery._data._endpoint_data import RequestMethod

content_type_by_day_interval_type = {
    DayIntervalType.INTER: ContentType.CUSTOM_INSTRUMENTS_INTERDAY_SUMMARIES,
    DayIntervalType.INTRA: ContentType.CUSTOM_INSTRUMENTS_INTRADAY_SUMMARIES,
}

# a20140be-3648-4892-9d1b-ce78ee8617fd
is_instrument_id = re.compile("[a-z0-9]{8}(-[a-z0-9]{4}){3}-[a-z0-9]{12}")


def get_query_params_from_kwargs(kwargs: dict, params: dict) -> list:
    query_params = []
    for param_name, request_param_name in params.items():
        value = kwargs.get(param_name)
        if value:
            query_params.append((request_param_name, value))
    return query_params


def get_params(params_details, data):
    retval = []
    for value in params_details:
        param_name = value.get("param_name")
        request_param_name = value.get("request_param_name")
        convert_function = value.get("convert_function")
        param = data.get(param_name)
        if not param:
            continue

        if convert_function:
            param = convert_function(param)
        retval.append((request_param_name, param))
    return retval


def get_content_type_by_interval(interval) -> ContentType:
    day_interval_type = get_day_interval_type(interval)
    return content_type_by_day_interval_type.get(day_interval_type)


# --------------------------------------------------------------------------------------
#   Response factory
# --------------------------------------------------------------------------------------


def custom_instruments_build_df(content_data: dict, **kwargs) -> pd.DataFrame:
    if isinstance(content_data, dict):
        content_data = [content_data]
    dataframe = pd.DataFrame(content_data)
    dataframe.fillna(pd.NA, inplace=True)
    return dataframe


def custom_instruments_intervals_build_df(content_data: dict, **kwargs) -> pd.DataFrame:
    data = content_data.get("data")
    headers = content_data.get("headers", [])
    columns = [header.get("name") for header in headers]
    dataframe = pd.DataFrame(data, columns=columns)
    convert_df_columns_to_datetime(dataframe, pattern="DATE", utc=True, delete_tz=True)
    dataframe.fillna(pd.NA, inplace=True)
    return dataframe


# --------------------------------------------------------------------------------------
#   Request factory
# --------------------------------------------------------------------------------------


def get_user_id(session=None) -> str:
    provider = DataProviderLayer(
        data_type=ContentType.CUSTOM_INSTRUMENTS_INSTRUMENTS,
        universe="S)IntrumentT.UUID",
    )
    provider._check_response = lambda *args, **kwargs: None
    response = provider.get_data(session=session)
    errors = response.errors
    messages = [error.message for error in errors]
    for message in messages:
        if message.startswith(
            "Validation Error: .UUID suffix UUID not matched with userID"
        ):
            _, user_id = message.rsplit(" ", 1)
            return user_id


def get_correct_symbol(symbol, session):
    # "MyNewInstrument"
    retval = symbol
    if not retval.startswith("S)"):
        retval = f"S){retval}"
    # "S)MyNewInstrument"
    if "." not in retval:
        user_id = get_user_id(session)
        retval = f"{retval}.{user_id}"
    # "S)MyNewInstrument.GE-1234"
    return retval


def convert_json_to_pandas(data: dict) -> pd.DataFrame:
    content_data = data.get("content_data")
    if isinstance(content_data, dict):
        content_data = [content_data]
    dataframe = pd.DataFrame(content_data)
    dataframe.fillna(pd.NA, inplace=True)
    return dataframe


def convert_historical_json_to_pandas(data: dict) -> pd.DataFrame:
    content_data = data.get("content_data", {})
    _data = content_data.get("data")
    headers = content_data.get("headers", [])
    columns = [header.get("name") for header in headers]
    dataframe = pd.DataFrame(_data, columns=columns)
    dataframe.fillna(pd.NA, inplace=True)
    return dataframe


class BaseRequestFactory(RequestFactory):
    def get_url(self, *args, **kwargs):
        if self.get_request_method(*args, **kwargs) == RequestMethod.POST:
            return super().get_url(*args, **kwargs)
        return super().get_url(*args, **kwargs) + "/{universe}"

    def get_path_parameters(self, *args, **kwargs):
        if self.get_request_method(*args, **kwargs) == RequestMethod.POST:
            return {}
        session = args[0]
        universe = kwargs.get("universe")
        if universe is None:
            raise RDError(-1, "universe can't be None")

        if not is_instrument_id.match(universe):
            universe = get_correct_symbol(universe, session)
        return {"universe": universe}


class CustomInstsRequestFactory(BaseRequestFactory):
    def get_body_parameters(self, *args, **kwargs):
        body_parameters = {}
        if self.get_request_method(*args, **kwargs) not in [
            RequestMethod.POST,
            RequestMethod.PUT,
        ]:
            return body_parameters

        symbol = kwargs.get("symbol")
        if symbol:
            if not symbol.startswith("S)"):
                symbol = f"S){symbol}"
            if "." not in symbol:
                session = args[0]
                user_id = get_user_id(session)
                symbol = f"{symbol}.{user_id}"
            body_parameters["symbol"] = symbol

        currency = kwargs.get("currency")
        if currency:
            body_parameters["currency"] = currency

        description = kwargs.get("description")
        if description:
            body_parameters["description"] = description

        exchange_name = kwargs.get("exchange_name")
        if exchange_name:
            body_parameters["exchangeName"] = exchange_name

        formula = kwargs.get("formula")
        if formula:
            body_parameters["formula"] = formula

        holidays = kwargs.get("holidays")
        if holidays:
            body_parameters["holidays"] = holidays

        instrument_name = kwargs.get("instrument_name")
        if instrument_name:
            body_parameters["instrumentName"] = instrument_name

        time_zone = kwargs.get("time_zone")
        if time_zone:
            body_parameters["timeZone"] = time_zone

        return body_parameters


# --------------------------------------------------------------------------------------
#   Raw data parser
# --------------------------------------------------------------------------------------


class CustomInstsParser(Parser):
    def parse_raw_response(self, raw_response, *args, **kwargs):
        parsed_data = {}
        is_success = False

        if raw_response is None:
            return is_success, parsed_data

        is_success = raw_response.status_code in success_http_codes + [
            requests.codes.no_content
        ]

        if is_success:
            parsed_data = self.process_successful_response(raw_response)

        else:
            parsed_data = self.process_failed_response(raw_response)

        return is_success, parsed_data

    def process_failed_response(self, raw_response):
        status = {
            "http_status_code": raw_response.status_code,
            "http_reason": get_response_reason(raw_response),
        }

        try:
            content_data = raw_response.json()
            content_error = content_data.get("error")

            if content_error:
                status["error"] = content_error
                error_code = content_error.get("code")
                if isinstance(error_code, str) and not error_code.isdigit():
                    error_code = raw_response.status_code
                error_message = content_error.get("message")
                errors = content_error.get("errors", {})
                errors = [error.get("reason") for error in errors if error]
                if errors:
                    errors = "\n".join(errors)
                    error_message = f"{error_message}: {errors}"
            elif "state" in content_data:
                state = content_data.get("state", {})
                error_code = state.get("code")
                data = content_data.get("data", [])
                reasons = [_data.get("reason", "") for _data in data]
                reason = "\n".join(reasons)
                error_message = f"{state.get('message')}: {reason}"
            else:
                error_code = raw_response.status_code
                error_message = raw_response.text

        except (TypeError, JSONDecodeError):
            error_code = raw_response.status_code
            error_message = raw_response.text

        if error_code == 403:
            if not error_message.endswith("."):
                error_message += ". "
            error_message += "Contact Refinitiv to check your permissions."

        parsed_data = {
            "status": status,
            "error_code": error_code,
            "error_message": error_message,
            "raw_response": raw_response,
        }

        return parsed_data


# --------------------------------------------------------------------------------------
#   Content data validator
# --------------------------------------------------------------------------------------


class CustomInstsContentValidator(ContentValidator):
    def validate(self, data, *args, **kwargs):
        is_valid = True
        content_data = data.get("content_data")
        status = data.get("status")
        status_code = status.get("http_status_code")

        if content_data is None and status_code != 204:
            is_valid = False
            data["error_code"] = 1
            data["error_message"] = "Content data is None"

        return is_valid


# --------------------------------------------------------------------------------------
#   Request factory
# --------------------------------------------------------------------------------------

interval_arg_parser = make_enum_arg_parser(Intervals, can_be_lower=True)


class CustomInstsSearchRequestFactory(RequestFactory):
    def get_query_parameters(self, *args, **kwargs):
        access = kwargs.get("access")
        return [
            ("access", access),
        ]


class CustomInstsEventsRequestFactory(BaseRequestFactory):
    def get_query_parameters(self, *args, **kwargs):
        query_params = get_params(
            [
                {
                    "param_name": "start",
                    "request_param_name": "start",
                    "convert_function": custom_insts_datetime_adapter.get_str,
                },
                {
                    "param_name": "end",
                    "request_param_name": "end",
                    "convert_function": custom_insts_datetime_adapter.get_str,
                },
                {"param_name": "count", "request_param_name": "count"},
            ],
            kwargs,
        )
        return query_params


class CustomInstsSummariesRequestFactory(BaseRequestFactory):
    def get_field_timestamp(self, *args, day_interval_type=None, **kwargs):
        return field_timestamp_by_day_interval_type.get(day_interval_type)

    def get_query_parameters(self, *args, **kwargs):
        query_params = get_params(
            [
                {
                    "param_name": "start",
                    "request_param_name": "start",
                    "convert_function": custom_insts_datetime_adapter.get_str,
                },
                {
                    "param_name": "end",
                    "request_param_name": "end",
                    "convert_function": custom_insts_datetime_adapter.get_str,
                },
                {"param_name": "count", "request_param_name": "count"},
            ],
            kwargs,
        )
        interval = kwargs.get("interval")
        if interval:
            interval = interval_arg_parser.get_str(interval)
            query_params.append(("interval", interval))
        return query_params


# --------------------------------------------------------------------------------------
#   Data provider
# --------------------------------------------------------------------------------------


class CustomInstsEventsDataProvider(HistoricalDataProvider):
    @staticmethod
    def _join_responses(
        responses: List[Tuple[str, "Response"]],
        fields: List[str],
        kwargs,
    ) -> "Response":
        axis_name = "Timestamp"
        return join_historical_responses(responses, axis_name, fields)


class CustomInstsSummariesDataProvider(HistoricalDataProvider):
    @staticmethod
    def _join_responses(
        responses: List[Tuple[str, "Response"]],
        fields: List[str],
        kwargs,
    ) -> "Response":
        axis_name = axis_by_day_interval_type.get(kwargs.get("day_interval_type"))
        return join_historical_responses(responses, axis_name, fields)


custom_instrument_data_provider = DataProvider(
    response=ResponseFactory(),
    request=CustomInstsRequestFactory(),
    parser=CustomInstsParser(),
    validator=ValidatorContainer(
        content_validator=CustomInstsContentValidator(),
        content_type_validator=ContentTypeValidator({"application/json", ""}),
    ),
)
custom_instrument_search_data_provider = DataProvider(
    request=CustomInstsSearchRequestFactory(),
    parser=CustomInstsParser(),
    validator=ValidatorContainer(
        content_validator=CustomInstsContentValidator(),
        content_type_validator=ContentTypeValidator({"application/json", ""}),
    ),
)

custom_instruments_events_data_provider = CustomInstsEventsDataProvider(
    request=CustomInstsEventsRequestFactory(),
)
custom_instruments_intraday_summaries_data_provider = CustomInstsSummariesDataProvider(
    request=CustomInstsSummariesRequestFactory(),
)
custom_instruments_interday_summaries_data_provider = CustomInstsSummariesDataProvider(
    request=CustomInstsSummariesRequestFactory(),
)



# coding: utf8
# contract_gen 2020-05-18 08:30:59.247816

__all__ = ["IndexConvexityAdjustmentType"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class IndexConvexityAdjustmentType(Enum):
    NONE = "None"
    BLACK_SCHOLES = "BlackScholes"
    REPLICATION = "Replication"
    LIBOR_SWAP_METHOD = "LiborSwapMethod"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(
            IndexConvexityAdjustmentType, _INDEX_CONVEXITY_ADJUSTMENT_TYPE_VALUES, some
        )

    @staticmethod
    def normalize(some):
        return _normalize(
            _INDEX_CONVEXITY_ADJUSTMENT_TYPE_VALUES_IN_LOWER_BY_INDEX_CONVEXITY_ADJUSTMENT_TYPE,
            some,
        )


_INDEX_CONVEXITY_ADJUSTMENT_TYPE_VALUES = tuple(
    t.value for t in IndexConvexityAdjustmentType
)
_INDEX_CONVEXITY_ADJUSTMENT_TYPE_VALUES_IN_LOWER_BY_INDEX_CONVEXITY_ADJUSTMENT_TYPE = {
    name.lower(): item
    for name, item in IndexConvexityAdjustmentType.__members__.items()
}



from enum import Enum, unique


@unique
class AssetClass(Enum):
    """
    Asset class values to build 'filter' parameter in request for SymbolConversion content object.
    """

    COMMODITIES = "Commodities"
    EQUITY_OR_INDEX_OPTIONS = "EquityOrIndexOptions"
    BOND_AND_STIR_FUTURES_AND_OPTIONS = "BondAndSTIRFuturesAndOptions"
    WARRANTS = "Warrants"
    EQUITIES = "Equities"
    INDICES = "Indices"
    EQUITY_INDEX_FUTURES = "EquityIndexFutures"
    FUNDS = "Funds"
    CERTIFICATES = "Certificates"
    BONDS = "Bonds"
    RESERVE_CONVERTIBLE = "ReverseConvertible"
    MINI_FUTURE = "MiniFuture"
    FX_AND_MONEY = "FXAndMoney"



from ._enums import Package
from .._content_provider import (
    UniverseContentValidator,
    ErrorParser,
)
from ..._tools import (
    universe_arg_parser,
    make_enum_arg_parser,
)
from ...delivery._data._data_provider import (
    DataProvider,
    RequestFactory,
    ValidatorContainer,
)


class EstimatesRequestFactory(RequestFactory):
    def get_query_parameters(self, *_, **kwargs) -> list:
        query_parameters = []
        universe = universe_arg_parser.get_str(kwargs.get("universe"), delim=",")
        query_parameters.append(("universe", universe))

        package = kwargs.get("package")
        if package is not None:
            package = package_estimates_arg_parser.get_str(package)
            query_parameters.append(("package", package))

        return query_parameters

    def extend_query_parameters(self, query_parameters, extended_params):
        # query_parameters -> [("param1", "val1"), ]
        result = dict(query_parameters)
        # result -> {"param1": "val1"}
        result.update(extended_params)
        # result -> {"param1": "val1", "extended_param": "value"}
        # return [("param1", "val1"), ("extended_param", "value")]
        return list(result.items())


package_estimates_arg_parser = make_enum_arg_parser(Package)

estimates_data_provider = DataProvider(
    request=EstimatesRequestFactory(),
    validator=ValidatorContainer(content_validator=UniverseContentValidator()),
    parser=ErrorParser(),
)



# coding: utf8

from enum import Enum, unique
from ..._enums._common_tools import _convert_to_str, _normalize


@unique
class CompoundingType(Enum):
    COMPOUNDED = "Compounded"
    CONTINUOUS = "Continuous"
    DISCOUNTED = "Discounted"
    MONEY_MARKET = "MoneyMarket"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(CompoundingType, _COMPOUNDING_TYPE_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_COMPOUNDING_TYPE_VALUES_IN_LOWER_BY_COMPOUNDING_TYPE, some)


_COMPOUNDING_TYPE_VALUES = tuple(t.value for t in CompoundingType)
_COMPOUNDING_TYPE_VALUES_IN_LOWER_BY_COMPOUNDING_TYPE = {
    name.lower(): item for name, item in CompoundingType.__members__.items()
}



# coding: utf8

from typing import Optional

from .._instrument_definition import InstrumentDefinition
from ._enums import (
    BusinessDayConvention,
    CdsConvention,
)
from ._models import (
    PremiumLegDefinition,
    ProtectionLegDefinition,
)


class CdsInstrumentDefinition(InstrumentDefinition):
    """
    API endpoint for Financial Contract analytics,
    that returns calculations relevant to each contract type.

    Parameters
    ----------
    instrument_tag : str, optional
        User defined string to identify the instrument.It can be used to link output
        results to the instrument definition. Only alphabetic, numeric and '- _.#=@'
        characters are supported. Optional.
    instrument_code : str, optional
        A cds RIC that is used to retrieve the description of the cds contract.
        Optional. If null, the protection_leg and the premium_leg  must be provided.
    cds_convention : CdsConvention, optional
        Define the cds convention. Possible values are:
        - 'ISDA' (start_date will default to accrued_begin_date, end_date will be
          adjusted to IMM Date),
        - 'UserDefined' (start_date will default to step_in_date, end_date will not be
          adjusted). Optional. Defaults to 'ISDA'.
    trade_date : str, optional
        The date the cds contract was created. Optional. By default the valuation date.
    step_in_date : str, optional
        The effective protection date. Optional. By default the trade_date + 1 calendar.
    start_date : str, optional
        The date the cds starts accruing interest. Its effective date. Optional. By
        default it is the accrued_begin_date (the last IMM date before trade_date) if
        cds_convention is ISDA, else it is the step_in_date.
    end_date : str, optional
        The maturity date of the cds contract. Mandatory if instrument_code is null.
        Either the end_date or the tenor must be provided.
    tenor : str, optional
        The period code that represents the time between the start date and end date the
        contract. Mandatory if instrument_code is null. Either the end_date or the tenor
        must be provided.
    start_date_moving_convention : BusinessDayConvention, optional
        The method to adjust the start_date. The possible values are:
        - ModifiedFollowing (adjusts dates according to the Modified Following
          convention - next business day unless is it goes into the next month,
          preceeding is used in that  case),
        - NextBusinessDay (adjusts dates according to the Following convention - Next
          Business Day),
        - PreviousBusinessDay (adjusts dates  according to the Preceeding convention -
          Previous Business Day),
        - NoMoving (does not adjust dates),
        - BbswModifiedFollowing (adjusts dates  according to the BBSW Modified Following
          convention). Optional. By default 'NoMoving' is used.
    end_date_moving_convention : BusinessDayConvention, optional
        The method to adjust the end_date.. The possible values are:
        - ModifiedFollowing (adjusts dates according to the Modified Following
          convention - next business day unless is it goes into the next month,
          preceeding is used in that  case),
        - NextBusinessDay (adjusts dates according to the Following convention - Next
          Business Day),
        - PreviousBusinessDay (adjusts dates  according to the Preceeding convention -
          Previous Business Day),
        - NoMoving (does not adjust dates),
        - BbswModifiedFollowing (adjusts dates  according to the BBSW Modified Following
          convention). Optional. By default 'NoMoving' is used.
    adjust_to_isda_end_date : bool, optional
        The way the end_date is adjusted if computed from tenor input.    The possible
        values are:
        - true ( the end_date is an IMM date computed from start_date according to ISDA
          rules, ),
        - false ( the end_date is computed from start_date according to
          end_dateMovingConvention), Optional. By default true is used if cds_convention
          is ISDA, else false is used.
    protection_leg : ProtectionLegDefinition, optional
        The Protection Leg of the CDS. It is the default leg. Mandatory if
        instrument_code is null. Optional if instrument_code not null.
    premium_leg : PremiumLegDefinition, optional
        The Premium Leg of the CDS. It is a swap leg paying a fixed coupon. Mandatory if
        instrument_code is null. Optional if instrument_code not null.
    accrued_begin_date : str, optional
        The last cashflow date. Optional. By default it is the last cashflow date.
    """

    def __init__(
        self,
        instrument_tag: Optional[str] = None,
        instrument_code: Optional[str] = None,
        cds_convention: Optional[CdsConvention] = None,
        trade_date: Optional[str] = None,
        step_in_date: Optional[str] = None,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
        tenor: Optional[str] = None,
        start_date_moving_convention: Optional[BusinessDayConvention] = None,
        end_date_moving_convention: Optional[BusinessDayConvention] = None,
        adjust_to_isda_end_date: Optional[bool] = None,
        protection_leg: Optional[ProtectionLegDefinition] = None,
        premium_leg: Optional[PremiumLegDefinition] = None,
        accrued_begin_date: Optional[str] = None,
    ) -> None:
        super().__init__()
        self.instrument_tag = instrument_tag
        self.instrument_code = instrument_code
        self.cds_convention = cds_convention
        self.trade_date = trade_date
        self.step_in_date = step_in_date
        self.start_date = start_date
        self.end_date = end_date
        self.tenor = tenor
        self.start_date_moving_convention = start_date_moving_convention
        self.end_date_moving_convention = end_date_moving_convention
        self.adjust_to_isda_end_date = adjust_to_isda_end_date
        self.protection_leg = protection_leg
        self.premium_leg = premium_leg
        self.accrued_begin_date = accrued_begin_date

    @classmethod
    def get_instrument_type(cls):
        return "Cds"

    @property
    def cds_convention(self):
        """
        Define the cds convention. Possible values are:
        - 'ISDA' (start_date will default to accrued_begin_date, end_date will be
          adjusted to IMM Date),
        - 'UserDefined' (start_date will default to step_in_date, end_date will not be
          adjusted). Optional. Defaults to 'ISDA'.
        :return: enum CdsConvention
        """
        return self._get_enum_parameter(CdsConvention, "cdsConvention")

    @cds_convention.setter
    def cds_convention(self, value):
        self._set_enum_parameter(CdsConvention, "cdsConvention", value)

    @property
    def end_date_moving_convention(self):
        """
        The method to adjust the end_date. The possible values are:
        - ModifiedFollowing (adjusts dates according to the Modified Following
          convention - next business day unless is it goes into the next month,
          preceeding is used in that  case),
        - NextBusinessDay (adjusts dates according to the Following convention - Next
          Business Day),
        - PreviousBusinessDay (adjusts dates  according to the Preceeding convention -
          Previous Business Day),
        - NoMoving (does not adjust dates),
        - BbswModifiedFollowing (adjusts dates  according to the BBSW Modified Following
          convention). Optional. By default 'NoMoving' is used.
        :return: enum BusinessDayConvention
        """
        return self._get_enum_parameter(
            BusinessDayConvention, "endDateMovingConvention"
        )

    @end_date_moving_convention.setter
    def end_date_moving_convention(self, value):
        self._set_enum_parameter(
            BusinessDayConvention, "endDateMovingConvention", value
        )

    @property
    def premium_leg(self):
        """
        The Premium Leg of the CDS. It is a swap leg paying a fixed coupon. Mandatory if
        instrument_code is null. Optional if instrument_code not null.
        :return: object PremiumLegDefinition
        """
        return self._get_object_parameter(PremiumLegDefinition, "premiumLeg")

    @premium_leg.setter
    def premium_leg(self, value):
        self._set_object_parameter(PremiumLegDefinition, "premiumLeg", value)

    @property
    def protection_leg(self):
        """
        The Protection Leg of the CDS. It is the default leg. Mandatory if
        instrument_code is null. Optional if instrument_code not null.
        :return: object ProtectionLegDefinition
        """
        return self._get_object_parameter(ProtectionLegDefinition, "protectionLeg")

    @protection_leg.setter
    def protection_leg(self, value):
        self._set_object_parameter(ProtectionLegDefinition, "protectionLeg", value)

    @property
    def start_date_moving_convention(self):
        """
        The method to adjust the start_date. The possible values are:
        - ModifiedFollowing (adjusts dates according to the Modified Following
          convention - next business day unless is it goes into the next month,
          preceeding is used in that  case),
        - NextBusinessDay (adjusts dates according to the Following convention - Next
          Business Day),
        - PreviousBusinessDay (adjusts dates  according to the Preceeding convention -
          Previous Business Day),
        - NoMoving (does not adjust dates),
        - BbswModifiedFollowing (adjusts dates  according to the BBSW Modified Following
          convention). Optional. By default 'NoMoving' is used.
        :return: enum BusinessDayConvention
        """
        return self._get_enum_parameter(
            BusinessDayConvention, "startDateMovingConvention"
        )

    @start_date_moving_convention.setter
    def start_date_moving_convention(self, value):
        self._set_enum_parameter(
            BusinessDayConvention, "startDateMovingConvention", value
        )

    @property
    def accrued_begin_date(self):
        """
        The last cashflow date. Optional. By default it is the last cashflow date.
        :return: str
        """
        return self._get_parameter("accruedBeginDate")

    @accrued_begin_date.setter
    def accrued_begin_date(self, value):
        self._set_parameter("accruedBeginDate", value)

    @property
    def adjust_to_isda_end_date(self):
        """
        The way the end_date is adjusted if computed from tenor input.    The possible
        values are:
        - true ( the end_date is an IMM date computed from start_date according to ISDA
          rules, ),
        - false ( the end_date is computed from start_date according to
          end_dateMovingConvention),
        Optional. By default true is used if cds_convention is ISDA, else false is used.
        :return: bool
        """
        return self._get_parameter("adjustToIsdaEndDate")

    @adjust_to_isda_end_date.setter
    def adjust_to_isda_end_date(self, value):
        self._set_parameter("adjustToIsdaEndDate", value)

    @property
    def end_date(self):
        """
        The maturity date of the cds contract. Mandatory if instrument_code is null.
        Either the end_date or the tenor must be provided.
        :return: str
        """
        return self._get_parameter("endDate")

    @end_date.setter
    def end_date(self, value):
        self._set_parameter("endDate", value)

    @property
    def instrument_code(self):
        """
        A cds RIC that is used to retrieve the description of the cds contract.
        Optional. If null, the protection_leg and the premium_leg  must be provided.
        :return: str
        """
        return self._get_parameter("instrumentCode")

    @instrument_code.setter
    def instrument_code(self, value):
        self._set_parameter("instrumentCode", value)

    @property
    def start_date(self):
        """
        The date the cds starts accruing interest. Its effective date. Optional. By
        default it is the accrued_begin_date (the last IMM date before trade_date) if
        cds_convention is ISDA, else it is the step_in_date.
        :return: str
        """
        return self._get_parameter("startDate")

    @start_date.setter
    def start_date(self, value):
        self._set_parameter("startDate", value)

    @property
    def step_in_date(self):
        """
        The effective protection date. Optional. By default the trade_date + 1 calendar.
        :return: str
        """
        return self._get_parameter("stepInDate")

    @step_in_date.setter
    def step_in_date(self, value):
        self._set_parameter("stepInDate", value)

    @property
    def tenor(self):
        """
        The period code that represents the time between the start date and end date the
        contract. Mandatory if instrument_code is null. Either the end_date or the tenor
        must be provided.
        :return: str
        """
        return self._get_parameter("tenor")

    @tenor.setter
    def tenor(self, value):
        self._set_parameter("tenor", value)

    @property
    def trade_date(self):
        """
        The date the cds contract was created. Optional. By default the valuation date.
        :return: str
        """
        return self._get_parameter("tradeDate")

    @trade_date.setter
    def trade_date(self, value):
        self._set_parameter("tradeDate", value)



# coding: utf8

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class AmortizationFrequency(Enum):
    EVERY12TH_COUPON = "Every12thCoupon"
    EVERY2ND_COUPON = "Every2ndCoupon"
    EVERY3RD_COUPON = "Every3rdCoupon"
    EVERY4TH_COUPON = "Every4thCoupon"
    EVERY_COUPON = "EveryCoupon"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(
            AmortizationFrequency, _AMORTIZATION_FREQUENCY_VALUES, some
        )

    @staticmethod
    def normalize(some):
        return _normalize(
            _AMORTIZATION_FREQUENCY_VALUES_IN_LOWER_BY_AMORTIZATION_FREQUENCY, some
        )


_AMORTIZATION_FREQUENCY_VALUES = tuple(t.value for t in AmortizationFrequency)
_AMORTIZATION_FREQUENCY_VALUES_IN_LOWER_BY_AMORTIZATION_FREQUENCY = {
    name.lower(): item for name, item in AmortizationFrequency.__members__.items()
}



# coding: utf8
# contract_gen 2020-05-18 08:30:59.218860

__all__ = ["InOrOut"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class InOrOut(Enum):
    IN = "In"
    OUT = "Out"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(InOrOut, _IN_OR_OUT_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_IN_OR_OUT_VALUES_IN_LOWER_BY_IN_OR_OUT, some)


_IN_OR_OUT_VALUES = tuple(t.value for t in InOrOut)
_IN_OR_OUT_VALUES_IN_LOWER_BY_IN_OR_OUT = {
    name.lower(): item for name, item in InOrOut.__members__.items()
}
